@@ -75,12 +75,13 @@ static bool thread_group_in_exec(Task* t) {
   return false;
 }
 
-KernelMapIterator::KernelMapIterator(Task* t) : tid(t->tid) {
+KernelMapIterator::KernelMapIterator(Task* t, bool* ok)
+  : tid(t->tid) {
   // See https://lkml.org/lkml/2016/9/21/423
   ASSERT(t, !thread_group_in_exec(t)) << "Task-group in execve, so reading "
                                          "/proc/.../maps may trigger kernel "
                                          "deadlock!";
-  init();
+  init(ok);
 }
 
 KernelMapIterator::~KernelMapIterator() {
@@ -89,11 +90,18 @@ KernelMapIterator::~KernelMapIterator() {
   }
 }
 
-void KernelMapIterator::init() {
+void KernelMapIterator::init(bool* ok) {
   char maps_path[PATH_MAX];
   sprintf(maps_path, "/proc/%d/maps", tid);
+  if (ok) {
+    *ok = true;
+  }
   if (!(maps_file = fopen(maps_path, "r"))) {
-    FATAL() << "Failed to open " << maps_path;
+    if (ok) {
+      *ok = false;
+    } else {
+      FATAL() << "Failed to open " << maps_path;
+    }
   }
   ++*this;
 }
@@ -171,7 +179,12 @@ void KernelMapIterator::operator++() {
 
 static KernelMapping read_kernel_mapping(pid_t tid, remote_ptr<void> addr) {
   MemoryRange range(addr, 1);
-  for (KernelMapIterator it(tid); !it.at_end(); ++it) {
+  bool ok;
+  KernelMapIterator it(tid, &ok);
+  if (!ok) {
+    return KernelMapping();
+  }
+  for (; !it.at_end(); ++it) {
     const KernelMapping& km = it.current();
     if (km.contains(range)) {
       return km;
@@ -209,9 +222,9 @@ AddressSpace::Mapping::Mapping(const KernelMapping& map,
     : map(map),
       recorded_map(recorded_map),
       emu_file(emu_file),
-      mapped_file_stat(move(mapped_file_stat)),
+      mapped_file_stat(std::move(mapped_file_stat)),
       local_addr(static_cast<uint8_t*>(local_addr)),
-      monitored_shared_memory(move(monitored)),
+      monitored_shared_memory(std::move(monitored)),
       flags(FLAG_NONE) {}
 
 static unique_ptr<struct stat> clone_stat(
@@ -242,8 +255,6 @@ AddressSpace::~AddressSpace() {
   session_->on_destroy(this);
 }
 
-void AddressSpace::after_clone() { allocate_watchpoints(); }
-
 static uint32_t find_offset_of_syscall_instruction_in(SupportedArch arch,
                                                       uint8_t* vdso_data,
                                                       size_t vdso_len) {
@@ -284,7 +295,7 @@ void AddressSpace::map_rr_page(AutoRemoteSyscalls& remote) {
   Task* t = remote.task();
   SupportedArch arch = t->arch();
 
-  const char *fname;
+  const char *fname = nullptr;
   switch (t->arch()) {
     case x86_64:
     case aarch64:
@@ -301,7 +312,8 @@ void AddressSpace::map_rr_page(AutoRemoteSyscalls& remote) {
 
   string path = find_helper_library(fname);
   if (path.empty()) {
-    FATAL() << "Failed to locate " << fname;
+    FATAL() << "Failed to locate " << fname << "; needed by "
+      << t->exe_path() << " (" << arch_name(t->arch()) << ")";
   }
   path += fname;
   size_t offset_pages = t->session().is_recording() ?
@@ -350,13 +362,15 @@ void AddressSpace::map_rr_page(AutoRemoteSyscalls& remote) {
       TRACED, PRIVILEGED, RECORDING_AND_REPLAY, t->arch());
 }
 
-void AddressSpace::unmap_all_but_rr_page(AutoRemoteSyscalls& remote) {
+void AddressSpace::unmap_all_but_rr_mappings(AutoRemoteSyscalls& remote,
+    UnmapOptions options) {
   vector<MemoryRange> unmaps;
   for (const auto& m : maps()) {
     // Do not attempt to unmap [vsyscall] --- it doesn't work.
     if (m.map.start() != AddressSpace::rr_page_start() &&
         m.map.start() != AddressSpace::preload_thread_locals_start() &&
-        !m.map.is_vsyscall()) {
+        !m.map.is_vsyscall() &&
+        (!options.exclude_vdso_vvar || (!m.map.is_vdso() && m.map.is_vvar()))) {
       unmaps.push_back(m.map);
     }
   }
@@ -752,7 +766,7 @@ KernelMapping AddressSpace::map(Task* t, remote_ptr<void> addr,
                                 unique_ptr<struct stat> mapped_file_stat,
                                 const KernelMapping* recorded_map,
                                 EmuFile::shr_ptr emu_file, void* local_addr,
-                                shared_ptr<MonitoredSharedMemory>&& monitored) {
+                                shared_ptr<MonitoredSharedMemory> monitored) {
   LOG(debug) << "mmap(" << addr << ", " << num_bytes << ", " << HEX(prot)
              << ", " << HEX(flags) << ", " << HEX(offset_bytes) << ")";
   num_bytes = ceil_page_size(num_bytes);
@@ -774,8 +788,10 @@ KernelMapping AddressSpace::map(Task* t, remote_ptr<void> addr,
   unmap_internal(t, addr, num_bytes);
 
   const KernelMapping& actual_recorded_map = recorded_map ? *recorded_map : m;
-  map_and_coalesce(t, m, actual_recorded_map, emu_file, move(mapped_file_stat),
-                   move(local_addr), move(monitored));
+  map_and_coalesce(t, m, actual_recorded_map, emu_file,
+                   std::move(mapped_file_stat),
+                   std::move(local_addr),
+                   std::move(monitored));
 
   // During an emulated exec, we will explicitly map in a (copy of) the VDSO
   // at the recorded address.
@@ -886,11 +902,10 @@ void AddressSpace::protect(Task* t, remote_ptr<void> addr, size_t num_bytes,
              << ")";
 
   MemoryRange last_overlap;
-  auto protector = [this, prot, &last_overlap](const Mapping& mm,
-                                               const MemoryRange& rem) {
+  auto protector = [this, prot, &last_overlap](Mapping m,
+                                               MemoryRange rem) {
     LOG(debug) << "  protecting (" << rem << ") ...";
 
-    Mapping m = move(mm);
     remove_from_map(m.map);
 
     // PROT_GROWSDOWN means that if this is a grows-down segment
@@ -917,7 +932,7 @@ void AddressSpace::protect(Task* t, remote_ptr<void> addr, size_t num_bytes,
           m.map.subrange(m.map.start(), rem.start()),
           m.recorded_map.subrange(m.recorded_map.start(), rem.start()),
           m.emu_file, clone_stat(m.mapped_file_stat), m.local_addr,
-          move(monitored));
+          std::move(monitored));
       underflow.flags = m.flags;
       add_to_map(underflow);
     }
@@ -982,7 +997,7 @@ void AddressSpace::fixup_mprotect_growsdown_parameters(Task* t) {
 
 void AddressSpace::remap(Task* t, remote_ptr<void> old_addr,
                          size_t old_num_bytes, remote_ptr<void> new_addr,
-                         size_t new_num_bytes) {
+                         size_t new_num_bytes, int flags) {
   LOG(debug) << "mremap(" << old_addr << ", " << old_num_bytes << ", "
              << new_addr << ", " << new_num_bytes << ")";
   old_num_bytes = ceil_page_size(old_num_bytes);
@@ -992,6 +1007,10 @@ void AddressSpace::remap(Task* t, remote_ptr<void> old_addr,
   KernelMapping km = mr.map.subrange(old_addr, min(mr.map.end(), old_addr + old_num_bytes));
 
   unmap_internal(t, old_addr, old_num_bytes);
+  if (flags & MREMAP_DONTUNMAP) {
+    // This can only ever be an anonymous private mapping.
+    map(t, old_addr, old_num_bytes, km.prot(), km.flags(), 0, string());
+  }
   if (0 == new_num_bytes) {
     return;
   }
@@ -1248,7 +1267,20 @@ bool AddressSpace::notify_watchpoint_fired(uintptr_t debug_status,
          watchpoint_triggered(debug_status,
                               it.second.debug_regs_for_exec_read);
     } else {
-      watchpoint_in_range = it.first.contains(hit_addr);
+      // The reported address may not match our watchpoint exactly.
+      // The ARM manual says:
+      //   The address recorded is within an address range of the size defined by the
+      //   DCZID_EL0.BS field. The start of the range is aligned to the size defined
+      //   by the DCZID_EL0.BS field and its end is not greater than the address that
+      //   triggered the watchpoint.
+      // So we construct a range spanning the whole block, then test that the range
+      // intersects a watchpoint range *and* that hit_addr is not past the first byte
+      // of the watched region.
+      auto block_size = dczid_el0_block_size();
+      auto slop = hit_addr.as_int() % block_size;
+      auto hit_range = MemoryRange(hit_addr - slop, block_size);
+      watchpoint_in_range = it.first.intersects(hit_range) &&
+        it.first.start() >= hit_addr;
     }
     if (write_triggered || read_triggered || exec_triggered || watchpoint_in_range) {
       it.second.changed = true;
@@ -1288,10 +1320,9 @@ void AddressSpace::unmap_internal(Task*, remote_ptr<void> addr,
                                   ssize_t num_bytes) {
   LOG(debug) << "munmap(" << addr << ", " << num_bytes << ")";
 
-  auto unmapper = [this](const Mapping& mm, const MemoryRange& rem) {
+  auto unmapper = [this](Mapping m, MemoryRange rem) {
     LOG(debug) << "  unmapping (" << rem << ") ...";
 
-    Mapping m = move(mm);
     remove_from_map(m.map);
 
     LOG(debug) << "  erased (" << m.map << ") ...";
@@ -1303,7 +1334,7 @@ void AddressSpace::unmap_internal(Task*, remote_ptr<void> addr,
       Mapping underflow(m.map.subrange(m.map.start(), rem.start()),
                         m.recorded_map.subrange(m.map.start(), rem.start()),
                         m.emu_file, clone_stat(m.mapped_file_stat),
-                        m.local_addr, move(monitored));
+                        m.local_addr, std::move(monitored));
       underflow.flags = m.flags;
       add_to_map(underflow);
     }
@@ -1514,9 +1545,9 @@ static void assert_segments_match(Task* t, const KernelMapping& input_m,
     err = "inodes differ";
   }
   if (err.size()) {
-    LOG(error) << "cached mmap:";
+    cerr << "cached mmap:" << endl;
     t->vm()->dump();
-    LOG(error) << "/proc/" << t->tid << "/mmaps:";
+    cerr << "/proc/" << t->tid << "/mmaps:" << endl;
     AddressSpace::print_process_maps(t);
     ASSERT(t, false) << "\nCached mapping " << m << " should be " << km << "; "
                      << err;
@@ -1531,12 +1562,11 @@ void AddressSpace::ensure_replay_matches_single_recorded_mapping(Task* t, Memory
   ASSERT(t, range.start() == floor_page_size(range.start()));
   ASSERT(t, range.end() == ceil_page_size(range.end()));
 
-  auto fixer = [this, t, range](const Mapping& mm, const MemoryRange&) {
-    if (mm.map == range) {
+  auto fixer = [this, t, range](Mapping mapping, MemoryRange) {
+    if (mapping.map == range) {
       // Existing single mapping covers entire range; nothing to do.
       return;
     }
-    Mapping mapping = move(mm);
 
     // These should be null during replay
     ASSERT(t, !mapping.mapped_file_stat);
@@ -1678,6 +1708,7 @@ AddressSpace::AddressSpace(Task* t, const string& exe, uint32_t exec_count)
   }
 }
 
+// Does not copy the task set; the new AddressSpace will be for new tasks.
 AddressSpace::AddressSpace(Session* session, const AddressSpace& o,
                            pid_t leader_tid, uint32_t leader_serial,
                            uint32_t exec_count)
@@ -1690,6 +1721,8 @@ AddressSpace::AddressSpace(Session* session, const AddressSpace& o,
       mem(o.mem),
       shm_sizes(o.shm_sizes),
       monitored_mem(o.monitored_mem),
+      dont_fork(o.dont_fork),
+      wipe_on_fork(o.wipe_on_fork),
       session_(session),
       vdso_start_addr(o.vdso_start_addr),
       monkeypatch_state(o.monkeypatch_state
@@ -1704,6 +1737,7 @@ AddressSpace::AddressSpace(Session* session, const AddressSpace& o,
       saved_auxv_(o.saved_auxv_),
       saved_interpreter_base_(o.saved_interpreter_base_),
       saved_ld_path_(o.saved_ld_path_),
+      last_free_memory(o.last_free_memory),
       first_run_event_(0) {
   for (auto& m : mem) {
     // The original address space continues to have exclusive ownership of
@@ -1734,7 +1768,7 @@ bool AddressSpace::post_vm_clone(Task* t) {
     return false;
   }
 
-  // Otherwise, the preload_thread_locals mapping is non-existent or ours.
+  // Otherwise, the preload_thread_locals mapping is nonexistent or ours.
   // Recreate it.
   AutoRemoteSyscalls remote(t);
   t->session().create_shared_mmap(remote, PRELOAD_THREAD_LOCALS_SIZE,
@@ -1771,7 +1805,13 @@ static vector<MemoryRange> split_range(const MemoryRange& range) {
 
 static void configure_watch_registers(vector<WatchConfig>& regs,
                                       const MemoryRange& range, WatchType type,
-                                      vector<int8_t>* assigned_regs) {
+                                      vector<int8_t>* assigned_regs,
+                                      AddressSpace::WatchpointAlignment alignment) {
+  if (alignment == AddressSpace::UNALIGNED) {
+    regs.push_back(WatchConfig(range.start(), range.size(), type));
+    return;
+  }
+
   // Zero-sized WatchConfigs return no ranges here, so are ignored.
   auto split_ranges = split_range(range);
 
@@ -1807,31 +1847,10 @@ static void configure_watch_registers(vector<WatchConfig>& regs,
   }
 }
 
-vector<WatchConfig> AddressSpace::get_watch_configs(
-    WillSetTaskState will_set_task_state) {
-  vector<WatchConfig> result;
-  for (auto& kv : watchpoints) {
-    vector<int8_t>* assigned_regs = nullptr;
-    if (will_set_task_state == SETTING_TASK_STATE) {
-      kv.second.debug_regs_for_exec_read.clear();
-      assigned_regs = &kv.second.debug_regs_for_exec_read;
-    }
-    const MemoryRange& r = kv.first;
-    int watching = kv.second.watched_bits();
-    if (EXEC_BIT & watching) {
-      configure_watch_registers(result, r, WATCH_EXEC, assigned_regs);
-    }
-    if (READ_BIT & watching) {
-      configure_watch_registers(result, r, WATCH_READWRITE, assigned_regs);
-    } else if (WRITE_BIT & watching) {
-      configure_watch_registers(result, r, WATCH_WRITE, nullptr);
-    }
-  }
-  return result;
-}
-
 vector<WatchConfig> AddressSpace::get_watchpoints_internal(
-    WatchpointFilter filter) {
+    WatchpointFilter filter,
+    WatchpointAlignment alignment,
+    UpdateWatchpointRegisterAssignments update_watchpoint_register_assignments) {
   vector<WatchConfig> result;
   for (auto& kv : watchpoints) {
     if (filter == CHANGED_WATCHPOINTS) {
@@ -1840,28 +1859,25 @@ vector<WatchConfig> AddressSpace::get_watchpoints_internal(
       }
       kv.second.changed = false;
     }
+    vector<int8_t>* assigned_regs = nullptr;
+    if (update_watchpoint_register_assignments == UPDATE_WATCHPOINT_REGISTER_ASSIGNMENTS) {
+      kv.second.debug_regs_for_exec_read.clear();
+      assigned_regs = &kv.second.debug_regs_for_exec_read;
+    }
     const MemoryRange& r = kv.first;
     int watching = kv.second.watched_bits();
     if (EXEC_BIT & watching) {
-      result.push_back(WatchConfig(r.start(), r.size(), WATCH_EXEC));
+      configure_watch_registers(result, r, WATCH_EXEC, assigned_regs, alignment);
     }
     if (READ_BIT & watching) {
-      result.push_back(WatchConfig(r.start(), r.size(), WATCH_READWRITE));
+      configure_watch_registers(result, r, WATCH_READWRITE, assigned_regs, alignment);
     } else if (WRITE_BIT & watching) {
-      result.push_back(WatchConfig(r.start(), r.size(), WATCH_WRITE));
+      configure_watch_registers(result, r, WATCH_WRITE, nullptr, alignment);
     }
   }
   return result;
 }
 
-vector<WatchConfig> AddressSpace::consume_watchpoint_changes() {
-  return get_watchpoints_internal(CHANGED_WATCHPOINTS);
-}
-
-vector<WatchConfig> AddressSpace::all_watchpoints() {
-  return get_watchpoints_internal(ALL_WATCHPOINTS);
-}
-
 bool AddressSpace::has_any_watchpoint_changes() {
   for (auto& kv : watchpoints) {
     if (kv.second.changed) {
@@ -1882,33 +1898,26 @@ bool AddressSpace::has_exec_watchpoint_fired(remote_code_ptr addr) {
 }
 
 bool AddressSpace::allocate_watchpoints() {
-  Task::DebugRegs regs = get_watch_configs(SETTING_TASK_STATE);
+  vector<WatchConfig> regs = get_watchpoints_internal(ALL_WATCHPOINTS, ALIGNED,
+      UPDATE_WATCHPOINT_REGISTER_ASSIGNMENTS);
 
-  if (regs.size() <= 0x7f) {
-    bool ok = true;
-    for (auto t : task_set()) {
-      if (!t->set_debug_regs(regs)) {
-        ok = false;
-      }
-    }
-    if (ok) {
-      return true;
-    }
+  if (task_set().empty()) {
+    // We can't validate the watchpoint set in this case
+    FATAL() << "No tasks???";
+  }
+  if ((*task_set().begin())->set_debug_regs(regs)) {
+    return true;
   }
 
-  regs.clear();
-  for (auto t2 : task_set()) {
-    t2->set_debug_regs(regs);
-  }
   for (auto kv : watchpoints) {
     kv.second.debug_regs_for_exec_read.clear();
   }
   return false;
 }
 
-static inline void assert_coalesceable(Task* t,
-                                       const AddressSpace::Mapping& lower,
-                                       const AddressSpace::Mapping& higher) {
+static inline void assert_coalescable(Task* t,
+                                      const AddressSpace::Mapping& lower,
+                                      const AddressSpace::Mapping& higher) {
   ASSERT(t, lower.emu_file == higher.emu_file);
   ASSERT(t, lower.flags == higher.flags);
   ASSERT(t,
@@ -1936,7 +1945,7 @@ void AddressSpace::coalesce_around(Task* t, MemoryMap::iterator it) {
       first_kv = next;
       break;
     }
-    assert_coalesceable(t, first_kv->second, next->second);
+    assert_coalescable(t, first_kv->second, next->second);
   }
   auto last_kv = it;
   while (true) {
@@ -1947,7 +1956,7 @@ void AddressSpace::coalesce_around(Task* t, MemoryMap::iterator it) {
       last_kv = prev;
       break;
     }
-    assert_coalesceable(t, prev->second, last_kv->second);
+    assert_coalescable(t, prev->second, last_kv->second);
   }
   ASSERT(t, last_kv != mem.end());
   if (first_kv == last_kv) {
@@ -2010,7 +2019,7 @@ void AddressSpace::maybe_update_breakpoints(Task* t, remote_ptr<uint8_t> addr,
 
 void AddressSpace::for_each_in_range(
     remote_ptr<void> addr, ssize_t num_bytes,
-    function<void(const Mapping& m, const MemoryRange& rem)> f, int how) {
+    function<void(Mapping m, MemoryRange rem)> f, int how) {
   remote_ptr<void> region_start = floor_page_size(addr);
   remote_ptr<void> last_unmapped_end = region_start;
   remote_ptr<void> region_end = ceil_page_size(addr + num_bytes);
@@ -2050,15 +2059,15 @@ void AddressSpace::for_each_in_range(
 void AddressSpace::map_and_coalesce(
     Task* t, const KernelMapping& m, const KernelMapping& recorded_map,
     EmuFile::shr_ptr emu_file, unique_ptr<struct stat> mapped_file_stat,
-    void* local_addr, shared_ptr<MonitoredSharedMemory>&& monitored) {
+    void* local_addr, shared_ptr<MonitoredSharedMemory> monitored) {
   LOG(debug) << "  mapping " << m;
 
   if (monitored) {
     monitored_mem.insert(m.start());
   }
   auto ins = mem.insert(MemoryMap::value_type(
-      m, Mapping(m, recorded_map, emu_file, move(mapped_file_stat), local_addr,
-                 move(monitored))));
+      m, Mapping(m, recorded_map, emu_file, std::move(mapped_file_stat),
+                 local_addr, std::move(monitored))));
   coalesce_around(t, ins.first);
 
   update_watchpoint_values(m.start(), m.end());
@@ -2131,7 +2140,7 @@ void AddressSpace::populate_address_space(Task* t) {
   ASSERT(t, found_stacks == 1);
 }
 
-static int random_addr_bits(SupportedArch arch) {
+static int addr_bits(SupportedArch arch) {
   switch (arch) {
     default:
       DEBUG_ASSERT(0 && "Unknown architecture");
@@ -2158,13 +2167,13 @@ static MemoryRange adjust_range_for_stack_growth(const KernelMapping& km) {
   return MemoryRange(start, km.end());
 }
 
-static bool overlaps_excluded_range(const RecordSession& session, MemoryRange range) {
+static MemoryRange overlaps_excluded_range(const RecordSession& session, MemoryRange range) {
   for (const auto& r : session.excluded_ranges()) {
     if (r.intersects(range)) {
-      return true;
+      return r;
     }
   }
-  return false;
+  return MemoryRange();
 }
 
 static bool is_all_memory_excluded(const RecordSession& session) {
@@ -2190,13 +2199,13 @@ static MemoryRange choose_global_exclusion_range(const RecordSession* session) {
 
   const uint64_t range_size = uint64_t(4)*1024*1024*1024*1024;
   while (true) {
-    int bits = random_addr_bits(x86_64);
+    int bits = addr_bits(x86_64);
     uint64_t r = ((uint64_t)(uint32_t)random() << 32) | (uint32_t)random();
     uint64_t r_addr = r & ((uint64_t(1) << bits) - 1);
     r_addr = min(r_addr, (uint64_t(1) << bits) - range_size);
     remote_ptr<void> addr = floor_page_size(remote_ptr<void>(r_addr));
     MemoryRange ret(addr, (uintptr_t)range_size);
-    if (!session || !overlaps_excluded_range(*session, ret)) {
+    if (!session || !overlaps_excluded_range(*session, ret).size()) {
       return ret;
     }
   }
@@ -2207,6 +2216,12 @@ MemoryRange AddressSpace::get_global_exclusion_range(const RecordSession* sessio
   return global_exclusion_range;
 }
 
+static remote_ptr<void> usable_address_space_end(Task* t) {
+  return remote_ptr<void>((uint64_t(1) << addr_bits(t->arch())) - page_size());
+}
+
+static const remote_ptr<void> addr_space_start(0x40000);
+
 remote_ptr<void> AddressSpace::chaos_mode_find_free_memory(RecordTask* t,
                                                            size_t len, remote_ptr<void> hint) {
   if (is_all_memory_excluded(t->session())) {
@@ -2223,101 +2238,135 @@ remote_ptr<void> AddressSpace::chaos_mode_find_free_memory(RecordTask* t,
     hint = nullptr;
   }
 
-  int bits = random_addr_bits(t->arch());
-  uint64_t addr_space_limit = uint64_t(1) << bits;
-  while (true) {
-    remote_ptr<void> addr;
-    if (hint) {
-      addr = hint;
-      // Don't try using the hint again.
-      hint = nullptr;
+  remote_ptr<void> start = hint;
+  if (!start) {
+    // Half the time, try to allocate at a completely random address. The other
+    // half of the time, we'll try to allocate immediately before or after a
+    // randomly chosen existing mapping.
+    if (random() % 2) {
+      uint64_t r = ((uint64_t)(uint32_t)random() << 32) | (uint32_t)random();
+      start = floor_page_size(remote_ptr<void>(r & ((uint64_t(1) << addr_bits(t->arch())) - 1)));
     } else {
-      // Half the time, try to allocate at a completely random address. The other
-      // half of the time, we'll try to allocate immediately before or after a
-      // randomly chosen existing mapping.
-      if (random() % 2) {
-        // Some of these addresses will not be mappable. That's fine, the
-        // kernel will fall back to a valid address if the hint is not valid.
-        uint64_t r = ((uint64_t)(uint32_t)random() << 32) | (uint32_t)random();
-        addr = floor_page_size(remote_ptr<void>(r & (addr_space_limit - 1)));
-      } else {
-        ASSERT(t, !mem.empty());
-        int map_index = random() % mem.size();
-        int map_count = 0;
-        for (const auto& m : maps()) {
-          if (map_count == map_index) {
-            addr = m.map.start();
-            break;
-          }
-          ++map_count;
+      ASSERT(t, !mem.empty());
+      int map_index = random() % mem.size();
+      int map_count = 0;
+      for (const auto& m : maps()) {
+        if (map_count == map_index) {
+          start = m.map.start();
+          break;
+        }
+        ++map_count;
+      }
+    }
+  }
+  // Reserve 3 pages at the end of userspace in case Monkeypatcher wants
+  // to allocate something there.
+  uint64_t reserve_area_for_monkeypatching = 3 * page_size();
+  remote_ptr<void> addr_space_end = usable_address_space_end(t) - reserve_area_for_monkeypatching;
+  // Clamp start so that we're in the usable address space.
+  start = max(start, addr_space_start);
+  start = min(start, addr_space_end - len);
+
+  // Search the address space in one direction all the way to the end,
+  // then in the other direction.
+  int direction = (random() % 2) ? 1 : -1;
+  remote_ptr<void> addr;
+  for (int iteration = 0; iteration < 2; ++iteration) {
+    // Invariant: [addr, addr+len) is always in the usable address space
+    // [addr_space_start, addr_space_end).
+    addr = start;
+    while (true) {
+      // Look for any reserved address space that overlaps [addr, addr+len]
+      // and store any overlapping range here. If multiple reserved areas
+      // overlap, we just pick one arbitrarily.
+      MemoryRange overlapping_range;
+      Maps m = maps_containing_or_after(addr);
+      if (m.begin() != m.end()) {
+        MemoryRange range = adjust_range_for_stack_growth(m.begin()->map);
+        if (range.start() < addr + len) {
+          overlapping_range = range;
         }
       }
-    }
-
-    // If there's a collision (which there always will be in the second case
-    // above), either move the mapping forwards or backwards in memory until it
-    // fits. Choose the direction randomly.
-    int direction = (random() % 2) ? 1 : -1;
-    while (true) {
-      Maps m = maps_starting_at(addr);
-      if (m.begin() == m.end()) {
-        break;
+      if (!overlapping_range.size()) {
+        MemoryRange r(addr, ceil_page_size(len));
+        if (r.intersects(rrpage_so_range)) {
+          overlapping_range = rrpage_so_range;
+        } else if (r.intersects(global_exclusion_range)) {
+          overlapping_range = global_exclusion_range;
+        } else if (!t->session().excluded_ranges().empty()) {
+          ASSERT(t, word_size(t->arch()) >= 8)
+            << "Chaos mode with ASAN/TSAN not supported in 32-bit processes";
+          MemoryRange excluded = overlaps_excluded_range(t->session(), r);
+          if (excluded.size()) {
+            overlapping_range = excluded;
+          }
+        }
       }
-      MemoryRange range = adjust_range_for_stack_growth(m.begin()->map);
-      if (range.start() >= addr + len) {
-        // No overlap with an existing mapping; we're good!
-        break;
+      if (!overlapping_range.size()) {
+        // No overlap and the range fits into our address space. Stop.
+        return addr;
       }
       if (direction == -1) {
-        addr = range.start() - len;
+        // Try moving backwards to allocate just before the start of
+        // the overlapping range.
+        if (overlapping_range.start() < addr_space_start + len) {
+          break;
+        }
+        addr = overlapping_range.start() - len;
       } else {
-        addr = range.end();
+        // Try moving forwards to allocate just after the end of
+        // the overlapping range.
+        if (overlapping_range.end() + len > addr_space_end) {
+          break;
+        }
+        addr = overlapping_range.end();
       }
     }
 
-    if (uint64_t(addr.as_int()) >= addr_space_limit ||
-        uint64_t(addr.as_int()) + ceil_page_size(len) >= addr_space_limit) {
-      // We fell off one end of the address space. Try everything again.
-      continue;
-    }
-    MemoryRange r(addr, ceil_page_size(len));
-    if (r.intersects(rrpage_so_range)) {
-      continue;
-    }
-    if (r.intersects(global_exclusion_range)) {
-      continue;
-    }
-    if (!t->session().excluded_ranges().empty()) {
-      ASSERT(t, word_size(t->arch()) >= 8)
-        << "Chaos mode with ASAN/TSAN not supported in 32-bit processes";
-      if (overlaps_excluded_range(t->session(), r)) {
-        continue;
-      }
-    }
-
-    return addr;
+    direction = -direction;
   }
+
+  return nullptr;
 }
 
-remote_ptr<void> AddressSpace::find_free_memory(size_t required_space,
-                                                remote_ptr<void> after) {
-  auto maps = maps_starting_at(after);
-  auto current = maps.begin();
-  while (current != maps.end()) {
-    auto next = current;
-    ++next;
-    if (next == maps.end()) {
-      if (current->map.end() + required_space >= current->map.end()) {
-        break;
+remote_ptr<void> AddressSpace::find_free_memory(Task* t,
+                                                size_t required_space,
+                                                remote_ptr<void> after,
+                                                FindFreeMemoryPolicy policy) {
+  if (after < last_free_memory &&
+      policy == FindFreeMemoryPolicy::USE_LAST_FREE_HINT) {
+    // Search for free memory starting at the last place we finished
+    // our search. This is more efficient than starting at the beginning
+    // every time.
+    after = last_free_memory;
+  }
+  remote_ptr<void> addr_space_end = usable_address_space_end(t);
+  ASSERT(t, required_space < UINT64_MAX - addr_space_end.as_int());
+
+  bool started_from_beginning = after.is_null();
+  while (true) {
+    auto maps = maps_starting_at(after);
+    auto current = maps.begin();
+    while (current != maps.end()) {
+      auto next = current;
+      ++next;
+      remote_ptr<void> end_of_free_space;
+      if (next == maps.end()) {
+        end_of_free_space = addr_space_end;
+      } else {
+        end_of_free_space = min(addr_space_end, next->map.start());
       }
-    } else {
-      if (current->map.end() + required_space <= next->map.start()) {
-        break;
+      if (current->map.end() + required_space <= end_of_free_space) {
+        return current->map.end();
       }
+      current = next;
     }
-    current = next;
+    if (started_from_beginning) {
+      return nullptr;
+    }
+    started_from_beginning = true;
+    after = addr_space_start;
   }
-  return current->map.end();
 }
 
 void AddressSpace::add_stap_semaphore_range(Task* task, MemoryRange range) {
@@ -2385,4 +2434,22 @@ void AddressSpace::fd_tables_changed() {
   }
 }
 
+bool AddressSpace::range_is_private_mapping(const MemoryRange& range) const {
+  MemoryRange r = range;
+  while (r.size() > 0) {
+    if (!has_mapping(r.start())) {
+      return false;
+    }
+    const AddressSpace::Mapping& m = mapping_of(r.start());
+    if (!(m.map.flags() & MAP_PRIVATE)) {
+      return false;
+    }
+    if (m.map.end() >= r.end()) {
+      return true;
+    }
+    r = MemoryRange(m.map.end(), r.end());
+  }
+  return true;
+}
+
 } // namespace rr
