@@ -48,7 +48,7 @@ public:
   /**
    * These are the flags we track internally to distinguish
    * between adjacent segments.  For example, the kernel
-   * considers a NORESERVE anonynmous mapping that's adjacent to
+   * considers a NORESERVE anonymous mapping that's adjacent to
    * a non-NORESERVE mapping distinct, even if all other
    * metadata are the same.  See |is_adjacent_mapping()|.
    */
@@ -231,7 +231,7 @@ enum DebugStatus {
 
 /**
  * A distinct watchpoint, corresponding to the information needed to
- * program a single x86 debug register.
+ * program a single hardware watchpoint.
  */
 struct WatchConfig {
   WatchConfig(remote_ptr<void> addr, size_t num_bytes, WatchType type)
@@ -239,6 +239,14 @@ struct WatchConfig {
   remote_ptr<void> addr;
   size_t num_bytes;
   WatchType type;
+
+  bool operator==(const WatchConfig& other) const {
+    return addr == other.addr && num_bytes == other.num_bytes &&
+      type == other.type;
+  }
+  bool operator!=(const WatchConfig& other) const {
+    return !(*this == other);
+  }
 };
 
 /**
@@ -302,13 +310,7 @@ public:
   typedef std::map<MemoryRange, Mapping, MappingComparator> MemoryMap;
   typedef std::shared_ptr<AddressSpace> shr_ptr;
 
-  ~AddressSpace();
-
-  /**
-   * Call this after a new task has been cloned within this
-   * address space.
-   */
-  void after_clone();
+  virtual ~AddressSpace();
 
   /**
    * Call this after a successful execve syscall has completed. At this point
@@ -428,7 +430,7 @@ public:
       std::unique_ptr<struct stat> mapped_file_stat = nullptr,
       const KernelMapping* recorded_map = nullptr,
       EmuFile::shr_ptr emu_file = nullptr, void* local_addr = nullptr,
-      std::shared_ptr<MonitoredSharedMemory>&& monitored = nullptr);
+      std::shared_ptr<MonitoredSharedMemory> monitored = nullptr);
 
   /**
    * Return the mapping and mapped resource for the byte at address 'addr'.
@@ -541,7 +543,7 @@ public:
    * [new_addr, old_addr + new_num_bytes), preserving metadata.
    */
   void remap(Task* t, remote_ptr<void> old_addr, size_t old_num_bytes,
-             remote_ptr<void> new_addr, size_t new_num_bytes);
+             remote_ptr<void> new_addr, size_t new_num_bytes, int flags);
 
   /**
    * Notify that data was written to this address space by rr or
@@ -582,7 +584,10 @@ public:
   void remove_watchpoint(remote_ptr<void> addr, size_t num_bytes,
                          WatchType type);
   void remove_all_watchpoints();
-  std::vector<WatchConfig> all_watchpoints();
+  std::vector<WatchConfig> all_watchpoints() {
+    return get_watchpoints_internal(ALL_WATCHPOINTS, UNALIGNED,
+      DONT_UPDATE_WATCHPOINT_REGISTER_ASSIGNMENTS);
+  }
 
   /**
    * Save all watchpoint state onto a stack.
@@ -620,7 +625,18 @@ public:
   /**
    * Return all changed watchpoints in |watches| and clear their changed flags.
    */
-  std::vector<WatchConfig> consume_watchpoint_changes();
+  std::vector<WatchConfig> consume_watchpoint_changes() {
+    return get_watchpoints_internal(CHANGED_WATCHPOINTS, UNALIGNED,
+      DONT_UPDATE_WATCHPOINT_REGISTER_ASSIGNMENTS);
+  }
+
+  /**
+   * Get hardware watchpoint assignments.
+   */
+  std::vector<WatchConfig> get_hw_watchpoints() {
+    return get_watchpoints_internal(ALL_WATCHPOINTS, ALIGNED,
+      DONT_UPDATE_WATCHPOINT_REGISTER_ASSIGNMENTS);
+  }
 
   void set_shm_size(remote_ptr<void> addr, size_t bytes) {
     shm_sizes[addr] = bytes;
@@ -658,6 +674,9 @@ public:
   ScopedFd& mem_fd() { return child_mem_fd; }
   void set_mem_fd(ScopedFd&& fd) { child_mem_fd = std::move(fd); }
 
+  ScopedFd& pagemap_fd() { return child_pagemap_fd; }
+  void set_pagemap_fd(ScopedFd&& fd) { child_pagemap_fd = std::move(fd); }
+
   Monkeypatcher& monkeypatcher() {
     DEBUG_ASSERT(monkeypatch_state);
     return *monkeypatch_state;
@@ -782,10 +801,18 @@ public:
 
   static uint32_t chaos_mode_min_stack_size() { return 8 * 1024 * 1024; }
 
-  /* Returns null if we should not override the memory allocation */
+  /* Returns null if we should return ENOMEM because there is no free space available. */
   remote_ptr<void> chaos_mode_find_free_memory(RecordTask* t, size_t len, remote_ptr<void> hint);
-  remote_ptr<void> find_free_memory(
-      size_t len, remote_ptr<void> after = remote_ptr<void>());
+  enum class FindFreeMemoryPolicy {
+    /* Use the first free memory after `after` */
+    STRICT_SEARCH,
+    /* Optimize for speed by starting the search from the address of the last
+       area returned by find_free_memory (if greater than `after`). */
+    USE_LAST_FREE_HINT,
+  };
+  remote_ptr<void> find_free_memory(Task* t,
+      size_t len, remote_ptr<void> after = remote_ptr<void>(),
+      FindFreeMemoryPolicy policy = FindFreeMemoryPolicy::STRICT_SEARCH);
 
   /**
    * The return value indicates whether we (re)created the preload_thread_locals
@@ -840,7 +867,12 @@ public:
   };
 
   void map_rr_page(AutoRemoteSyscalls& remote);
-  void unmap_all_but_rr_page(AutoRemoteSyscalls& remote);
+  struct UnmapOptions {
+    bool exclude_vdso_vvar;
+    UnmapOptions() : exclude_vdso_vvar(false) {}
+  };
+  void unmap_all_but_rr_mappings(AutoRemoteSyscalls& remote,
+                                 UnmapOptions options = UnmapOptions());
 
   void erase_task(Task* t) {
     this->HasTaskSet::erase_task(t);
@@ -858,6 +890,13 @@ public:
 
   static MemoryRange get_global_exclusion_range(const RecordSession* session);
 
+  // Whether to return WatchConfigs consisting of only aligned locations
+  // suitable for hardware watchpoint registers.
+  enum WatchpointAlignment { UNALIGNED, ALIGNED };
+
+  // Returns true if the range is completely covered by private mappings
+  bool range_is_private_mapping(const MemoryRange& range) const;
+
 private:
   struct Breakpoint;
   typedef std::map<remote_code_ptr, Breakpoint> BreakpointMap;
@@ -887,12 +926,17 @@ private:
   bool update_watchpoint_value(const MemoryRange& range,
                                Watchpoint& watchpoint);
   void update_watchpoint_values(remote_ptr<void> start, remote_ptr<void> end);
+  // Whether to handle all watchpoints or just data watchpoints whose data
+  // has changed. In the latter case we clear their changed status.
   enum WatchpointFilter { ALL_WATCHPOINTS, CHANGED_WATCHPOINTS };
-  std::vector<WatchConfig> get_watchpoints_internal(WatchpointFilter filter);
-
-  enum WillSetTaskState { SETTING_TASK_STATE, NOT_SETTING_TASK_STATE };
-  std::vector<WatchConfig> get_watch_configs(
-      WillSetTaskState will_set_task_state);
+  // Whether to update the watchpoint's assigned register list. Use
+  // UPDATE_WATCHPOINT_REGISTER_ASSIGNMENTS when we'll use the watchpoints
+  // to configure HW watchpoint registers.
+  enum UpdateWatchpointRegisterAssignments { UPDATE_WATCHPOINT_REGISTER_ASSIGNMENTS,
+      DONT_UPDATE_WATCHPOINT_REGISTER_ASSIGNMENTS };
+  std::vector<WatchConfig> get_watchpoints_internal(WatchpointFilter filter,
+        WatchpointAlignment alignment,
+        UpdateWatchpointRegisterAssignments update_watchpoint_register_assignments);
 
   /**
    * Construct a minimal set of watchpoints to be enabled based
@@ -924,11 +968,14 @@ private:
    * Pass |ITERATE_CONTIGUOUS| to stop iterating when the last
    * contiguous mapping after |addr| within the region is seen.
    * Default is to iterate all mappings in the region.
+   *
+   * The callback takes parameters by value to avoid dangling
+   * references if the memory map is modified inside the callback.
    */
   enum { ITERATE_DEFAULT, ITERATE_CONTIGUOUS };
   void for_each_in_range(
       remote_ptr<void> addr, ssize_t num_bytes,
-      std::function<void(const Mapping& m, const MemoryRange& rem)> f,
+      std::function<void(Mapping m, MemoryRange rem)> f,
       int how = ITERATE_DEFAULT);
 
   /**
@@ -940,7 +987,7 @@ private:
                         EmuFile::shr_ptr emu_file,
                         std::unique_ptr<struct stat> mapped_file_stat,
                         void* local_addr,
-                        std::shared_ptr<MonitoredSharedMemory>&& monitored);
+                        std::shared_ptr<MonitoredSharedMemory> monitored);
 
   void remove_from_map(const MemoryRange& range) {
     mem.erase(range);
@@ -1086,9 +1133,9 @@ private:
   /* Path of the real executable image this address space was
    * exec()'d with. */
   std::string exe;
-  /* Path of the intepreter, if any, of exe. */
+  /* Path of the interpreter, if any, of exe. */
   std::string interp_name_;
-  /* Base address of the interpeter (might be null!) */
+  /* Base address of the interpreter (might be null!) */
   remote_ptr<void> interp_base_;
   /* Pid of first task for this address space */
   pid_t leader_tid_;
@@ -1148,6 +1195,8 @@ private:
   remote_ptr<void> saved_interpreter_base_;
   std::string saved_ld_path_;
 
+  remote_ptr<void> last_free_memory;
+
   /**
    * The time of the first event that ran code for a task in this address space.
    * 0 if no such event has occurred.
@@ -1164,13 +1213,15 @@ private:
 
   /**
    * Ensure that the cached mapping of |t| matches /proc/maps,
-   * using adjancent-map-merging heuristics that are as lenient
+   * using adjacent-map-merging heuristics that are as lenient
    * as possible given the data available from /proc/maps.
    */
   static void check_segment_iterator(void* vasp, Task* t,
                                      const struct map_iterator_data* data);
 
   AddressSpace operator=(const AddressSpace&) = delete;
+
+  ScopedFd child_pagemap_fd;
 };
 
 /**
@@ -1179,8 +1230,8 @@ private:
  */
 class KernelMapIterator {
 public:
-  KernelMapIterator(Task* t);
-  KernelMapIterator(pid_t tid) : tid(tid) { init(); }
+  KernelMapIterator(Task* t, bool* ok = nullptr);
+  KernelMapIterator(pid_t tid, bool* ok = nullptr) : tid(tid) { init(ok); }
   ~KernelMapIterator();
 
   // It's very important to keep in mind that btrfs files can have the wrong
@@ -1195,7 +1246,7 @@ public:
   void operator++();
 
 private:
-  void init();
+  void init(bool* ok = nullptr);
 
   pid_t tid;
   FILE* maps_file;
