@@ -7,6 +7,8 @@
 #include <sys/socket.h>
 #include <sys/types.h>
 
+#include <sstream>
+
 #include "rr/rr.h"
 
 #include "RecordSession.h"
@@ -15,6 +17,7 @@
 #include "Session.h"
 #include "Task.h"
 #include "core.h"
+#include "kernel_abi.h"
 #include "kernel_metadata.h"
 #include "log.h"
 #include "util.h"
@@ -42,6 +45,11 @@ void AutoRestoreMem::init(const void* mem, ssize_t num_bytes) {
 
   remote.regs().set_sp(remote.regs().sp() - len);
   remote.task()->set_regs(remote.regs());
+  if (remote.task()->is_exiting()) {
+    // Leave addr == nullptr
+    return;
+  }
+
   addr = remote.regs().sp();
 
   data.resize(len);
@@ -238,8 +246,12 @@ void AutoRemoteSyscalls::maybe_fix_stack_pointer() {
 AutoRemoteSyscalls::~AutoRemoteSyscalls() { restore_state_to(t); }
 
 void AutoRemoteSyscalls::restore_state_to(Task* t) {
+  // Check if the task was unexpectedly killed via SIGKILL or equivalent.
+  bool is_exiting = !t->is_stopped() || t->ptrace_event() == PTRACE_EVENT_EXIT ||
+    t->was_reaped();
+
   // Unmap our scatch region if required
-  if (scratch_mem_was_mapped) {
+  if (scratch_mem_was_mapped && !is_exiting) {
     AutoRemoteSyscalls remote(t, DISABLE_MEMORY_PARAMS);
     remote.infallible_syscall(syscall_number_for_munmap(arch()),
                               fixed_sp - 4096, 4096);
@@ -252,6 +264,14 @@ void AutoRemoteSyscalls::restore_state_to(Task* t) {
   auto regs = initial_regs;
   regs.set_ip(initial_ip);
   regs.set_sp(initial_sp);
+  if (is_exiting) {
+    // Don't restore status; callers need to see the task is exiting.
+    // And the other stuff we don't below won't work.
+    // But do restore registers so it looks like the exit happened in a clean state.
+    t->set_regs(regs);
+    return;
+  }
+
   if (t->arch() == aarch64 && regs.syscall_may_restart()) {
     // On AArch64, the kernel restarts aborted syscalls using an internal `orig_x0`.
     // This gets overwritten everytime we make a syscall so we need to restore it
@@ -267,7 +287,11 @@ void AutoRemoteSyscalls::restore_state_to(Task* t) {
     regs.set_arg1(regs.orig_arg1());
     t->set_regs(regs);
     if (t->enter_syscall(true)) {
-      t->resume_execution(RESUME_SYSCALL, RESUME_WAIT, RESUME_NO_TICKS);
+      if (!t->resume_execution(RESUME_SYSCALL, RESUME_WAIT_NO_EXIT, RESUME_NO_TICKS)) {
+        // Tracee died unexpectedly, there is nothing more we can do.
+        // Do not restore the status, we want callers to see that the task died.
+        return;
+      }
     }
     regs.set_ip(initial_ip);
     regs.set_syscallno(regs.original_syscallno());
@@ -283,7 +307,11 @@ void AutoRemoteSyscalls::restore_state_to(Task* t) {
     t->set_regs(regs);
     RecordTask* rt = static_cast<RecordTask*>(t);
     while (true) {
-      rt->resume_execution(RESUME_CONT, RESUME_WAIT, RESUME_NO_TICKS);
+      if (!rt->resume_execution(RESUME_CONT, RESUME_WAIT_NO_EXIT, RESUME_NO_TICKS)) {
+        // Tracee died unexpectedly, there is nothing more we can do.
+        // Do not restore the status, we want callers to see that the task died.
+        return;
+      }
       if (rt->ptrace_event())
         break;
       rt->stash_sig();
@@ -303,6 +331,8 @@ void AutoRemoteSyscalls::restore_state_to(Task* t) {
     // needs to be able to interrupt re-startable system calls, it is required
     // to set TIF_SIGPENDING, but the fact that this works is of course a very
     // deep implementation detail.
+    // If this fails then the tracee must be dead or no longer traced, in which
+    // case we no longer care about its TIF_SIGPENDING status.
     t->do_ptrace_interrupt();
   }
 }
@@ -323,15 +353,23 @@ static bool ignore_signal(Task* t) {
     }
     return true;
   }
-  ASSERT(t, false) << "Unexpected signal " << signal_name(sig);
+  siginfo_t siginfo;
+  errno = 0;
+  t->fallible_ptrace(PTRACE_GETSIGINFO, nullptr, &siginfo);
+  if (errno) {
+    ASSERT(t, false) << "Unexpected signal " << signal_name(sig);
+  } else {
+    ASSERT(t, false) << "Unexpected signal " << siginfo;
+  }
   return false;
 }
 
 long AutoRemoteSyscalls::syscall_base(int syscallno, Registers& callregs) {
   LOG(debug) << "syscall " << syscall_name(syscallno, t->arch()) << " " << callregs;
 
-  if (t->is_dying()) {
+  if (t->is_exiting()) {
     LOG(debug) << "Task is dying, don't try anything.";
+    ASSERT(t, t->stopped_or_unexpected_exit()) << "Already seen exit event";
     return -ESRCH;
   }
 
@@ -353,17 +391,17 @@ long AutoRemoteSyscalls::syscall_base(int syscallno, Registers& callregs) {
   bool from_seccomp = initial_at_seccomp && t->ptrace_event() == PTRACE_EVENT_SECCOMP;
   if (use_singlestep_path && !from_seccomp) {
     while (true) {
-      t->resume_execution(RESUME_SINGLESTEP, RESUME_WAIT, RESUME_NO_TICKS);
+      if (!t->resume_execution(RESUME_SINGLESTEP, RESUME_WAIT_NO_EXIT, RESUME_NO_TICKS)) {
+        // Tracee was killed, there is nothing more we can do.
+        ASSERT(t, t->stopped_or_unexpected_exit()) << "Couldn't singlestep";
+        return -ESRCH;
+      }
       LOG(debug) << "Used singlestep path; status=" << t->status();
       // When a PTRACE_EVENT_EXIT is returned we don't update registers
       if (t->ip() != callregs.ip()) {
         // We entered the syscall, so stop now
         break;
       }
-      if (t->ptrace_event() == PTRACE_EVENT_EXIT) {
-        // We died, just let it be
-        break;
-      }
       if (t->stop_sig() == SIGTRAP && t->get_siginfo().si_code == TRAP_TRACE) {
         // On aarch64, if we were previously in a syscall-exit stop, continuing
         // with PTRACE_SINGLESTEP will result in incurring a trap upon execution
@@ -377,25 +415,26 @@ long AutoRemoteSyscalls::syscall_base(int syscallno, Registers& callregs) {
       ASSERT(t, false) << "Unexpected status " << t->status();
     }
   } else {
-    bool exited = false;
     if (from_seccomp) {
       LOG(debug) << "Skipping enter_syscall - already at seccomp stop";
     } else {
-      exited = !t->enter_syscall(true);
+      if (!t->enter_syscall(true)) {
+        // Tracee was killed, there is nothing more we can do.
+        // Ensure callers see the task death status.
+        ASSERT(t, t->stopped_or_unexpected_exit()) << "couldn't enter syscall";
+        return -ESRCH;
+      }
       LOG(debug) << "Used enter_syscall; status=" << t->status();
     }
-    // proceed to syscall exit
-    if (!exited) {
-      t->resume_execution(RESUME_SYSCALL, RESUME_WAIT, RESUME_NO_TICKS);
-      LOG(debug) << "syscall exit status=" << t->status();
+    if (!t->resume_execution(RESUME_SYSCALL, RESUME_WAIT_NO_EXIT, RESUME_NO_TICKS)) {
+      // Tracee was killed, there is nothing more we can do.
+      // Ensure callers see the task death status.
+      ASSERT(t, t->stopped_or_unexpected_exit()) << "couldn't resume syscall";
+      return -ESRCH;
     }
+    LOG(debug) << "syscall exit status=" << t->status();
   }
   while (true) {
-    // If the syscall caused the task to exit, just stop now with that status.
-    if (t->ptrace_event() == PTRACE_EVENT_EXIT) {
-      restore_wait_status = t->status();
-      break;
-    }
     if (t->status().is_syscall() ||
         (t->stop_sig() == SIGTRAP &&
          is_kernel_trap(t->get_siginfo().si_code))) {
@@ -405,19 +444,28 @@ long AutoRemoteSyscalls::syscall_base(int syscallno, Registers& callregs) {
     }
     if (is_clone_syscall(syscallno, t->arch()) &&
         t->clone_syscall_is_complete(&new_tid_, t->arch())) {
-      t->resume_execution(RESUME_SYSCALL, RESUME_WAIT, RESUME_NO_TICKS);
+      if (!t->resume_execution(RESUME_SYSCALL, RESUME_WAIT_NO_EXIT, RESUME_NO_TICKS)) {
+        // Tracee was killed, there is nothing more we can do.
+        ASSERT(t, t->stopped_or_unexpected_exit()) << "Couldn't resume clone";
+        return -ESRCH;
+      }
       LOG(debug) << "got clone event; new status=" << t->status();
       continue;
     }
     if (ignore_signal(t)) {
       if (t->regs().syscall_may_restart()) {
         if (!t->enter_syscall(true)) {
-          // We died, just let it be
-          break;
+          // Tracee was killed, there is nothing more we can do.
+          ASSERT(t, t->stopped_or_unexpected_exit()) << "Couldn't restart";
+          return -ESRCH;
         }
         LOG(debug) << "signal ignored; restarting syscall, status="
                    << t->status();
-        t->resume_execution(RESUME_SYSCALL, RESUME_WAIT, RESUME_NO_TICKS);
+        if (!t->resume_execution(RESUME_SYSCALL, RESUME_WAIT_NO_EXIT, RESUME_NO_TICKS)) {
+          // Tracee was killed, there is nothing more we can do.
+          ASSERT(t, t->stopped_or_unexpected_exit()) << "Couldn't resume restart";
+          return -ESRCH;
+        }
         LOG(debug) << "syscall exit status=" << t->status();
         continue;
       }
@@ -430,13 +478,8 @@ long AutoRemoteSyscalls::syscall_base(int syscallno, Registers& callregs) {
     break;
   }
 
-  if (t->is_dying()) {
-    LOG(debug) << "Task is dying, no status result";
-    return -ESRCH;
-  } else {
-    LOG(debug) << "done, result=" << t->regs().syscall_result();
-    return t->regs().syscall_result();
-  }
+  LOG(debug) << "done, result=" << t->regs().syscall_result();
+  return t->regs().syscall_result();
 }
 
 SupportedArch AutoRemoteSyscalls::arch() const { return t->arch(); }
@@ -543,7 +586,8 @@ static long child_recvmsg(AutoRemoteSyscalls& remote, int child_sock) {
     sizeof(msg), &msg, &ok);
 
   if (!ok) {
-    ASSERT(remote.task(), errno == ESRCH);
+    ASSERT(remote.task(), errno == ESRCH) << "Error writing " << remote_buf.get()
+        << " in " << remote.task()->tid;
     LOG(debug) << "Failed to write memory";
     return -ESRCH;
   }
@@ -566,23 +610,43 @@ static long child_recvmsg(AutoRemoteSyscalls& remote, int child_sock) {
   return their_fd;
 }
 
-static int recvmsg_socket(ScopedFd& sock, bool blocking=true) {
-  fd_message<NativeArch> msg;
-  struct msghdr *msgp = (struct msghdr*)&msg.msg;
-  int flag = MSG_CMSG_CLOEXEC;
+#define MAX_FDS_READ 2
+
+// Try to read a single-character message from `sock`. Will collect
+// up to MAX_FDS_READ fds in an SCM_RIGHTS control message and return those
+// fds. Returns an empty vector if reading the message fails.
+static vector<ScopedFd> maybe_receive_fds(ScopedFd& sock, bool blocking = true) {
+  vector<ScopedFd> ret;
+  struct msghdr msg;
+  memset(&msg, 0, sizeof(msg));
+  char ch;
+  struct iovec iov = { &ch, 1 };
+  msg.msg_iov = &iov;
+  msg.msg_iovlen = 1;
+  char cmsgbuf[(CMSG_SPACE(MAX_FDS_READ * sizeof(int)))];
+  msg.msg_control = cmsgbuf;
+  msg.msg_controllen = sizeof(cmsgbuf);
+  int flags = MSG_CMSG_CLOEXEC;
   if (!blocking) {
-    flag |= MSG_DONTWAIT;
+    flags |= MSG_DONTWAIT;
   }
-  if (0 > recvmsg(sock, msgp, flag)) {
-    return -1;
+  if (recvmsg(sock, &msg, flags) < 0) {
+    return ret;
   }
 
-  struct cmsghdr* cmsg = CMSG_FIRSTHDR(msgp);
-  DEBUG_ASSERT(cmsg && cmsg->cmsg_level == SOL_SOCKET &&
-               cmsg->cmsg_type == SCM_RIGHTS);
-  int our_fd = *(int*)CMSG_DATA(cmsg);
-  DEBUG_ASSERT(our_fd >= 0);
-  return our_fd;
+  struct cmsghdr* cmsg = CMSG_FIRSTHDR(&msg);
+  if (!cmsg || cmsg->cmsg_level != SOL_SOCKET ||
+      cmsg->cmsg_type != SCM_RIGHTS) {
+    FATAL() << "Invalid cmsg";
+  }
+  int num_fds = (cmsg->cmsg_len - CMSG_LEN(0))/sizeof(int);
+  for (int i = 0; i < num_fds; i++) {
+    int fd;
+    memcpy(&fd, CMSG_DATA(cmsg) + i*sizeof(int), sizeof(int));
+    DEBUG_ASSERT(fd >= 0);
+    ret.push_back(ScopedFd(fd));
+  }
+  return ret;
 }
 
 static void sendmsg_socket(ScopedFd& sock, int fd_to_send)
@@ -601,17 +665,49 @@ static void sendmsg_socket(ScopedFd& sock, int fd_to_send)
   }
 }
 
+static Task* thread_group_leader_for_fds(Task* t) {
+  for (Task* tt : t->fd_table()->task_set()) {
+    if (tt->tgid() == tt->rec_tid && !tt->seen_ptrace_exit_event()) {
+      return tt;
+    }
+  }
+  return nullptr;
+}
+
 template <typename Arch> ScopedFd AutoRemoteSyscalls::retrieve_fd_arch(int fd) {
+  ScopedFd ret;
+  if (!pid_fd.is_open()) {
+    // Try to use pidfd_getfd to get the fd without round-tripping to the tracee.
+    // pidfd_getfd requires a threadgroup leader, so find one if we can.
+    Task* tg_leader_for_fds = thread_group_leader_for_fds(t);
+    if (tg_leader_for_fds) {
+      pid_fd = ScopedFd(::syscall(NativeArch::pidfd_open, tg_leader_for_fds->tid, 0));
+      ASSERT(t, pid_fd.is_open() || errno == ENOSYS)
+        << "Error in pidfd_open errno=" << errno_name(errno);
+    }
+  }
+  if (pid_fd.is_open()) {
+    ret = ScopedFd(::syscall(NativeArch::pidfd_getfd, pid_fd.get(), fd, 0));
+    if (ret.is_open()) {
+      return ret;
+    }
+    ASSERT(t, errno == ENOSYS) << "Failed in pidfd_getfd errno=" << errno_name(errno);
+  }
+
+  // Clear out any pending message in the socket.
+  maybe_receive_fds(task()->session().tracee_socket_receiver_fd(), false);
+
   long child_syscall_result =
       child_sendmsg<Arch>(*this, task()->session().tracee_fd_number(), fd);
   if (child_syscall_result == -ESRCH) {
-    return ScopedFd();
+    return ret;
   }
   ASSERT(t, child_syscall_result > 0) << "Failed to sendmsg() in tracee; err="
                                       << errno_name(-child_syscall_result);
-  int our_fd = recvmsg_socket(task()->session().tracee_socket_fd());
-  ASSERT(t, our_fd >= 0) << "Failed to receive fd";
-  return ScopedFd(our_fd);
+  vector<ScopedFd> fds = maybe_receive_fds(task()->session().tracee_socket_fd());
+  ASSERT(t, !fds.empty()) << "Failed to receive fd";
+  ASSERT(t, fds.size() == 1);
+  return std::move(fds[0]);
 }
 
 ScopedFd AutoRemoteSyscalls::retrieve_fd(int fd) {
@@ -623,28 +719,21 @@ template <typename Arch> int AutoRemoteSyscalls::send_fd_arch(const ScopedFd &ou
     return -EBADF;
   }
 
+  // Clear out any pending message from the socket.
+  maybe_receive_fds(task()->session().tracee_socket_receiver_fd(), false);
+
   LOG(debug) << "Sending fd " << our_fd.get() << " via socket fd " << task()->session().tracee_socket_fd().get();
   sendmsg_socket(task()->session().tracee_socket_fd(), our_fd.get());
 
   long child_syscall_result =
       child_recvmsg<Arch>(*this, task()->session().tracee_fd_number());
-  if (child_syscall_result == -ESRCH) {
-    /* The child did not receive the message. Read it out of the socket
-     * buffer so it doesn't get read by another child later!
-     * Note that if the child was killed, we might get an error
-     * (i.e. process exit stop before syscall exit stop) even
-     * if the fd is already removed from the socket.
-     * Because of this, we use a non-blocking recvmsg to avoid blocking forever.
-     */
-    int fd = recvmsg_socket(task()->session().tracee_socket_receiver_fd(), false);
-    if (fd >= 0) {
-      ASSERT(t, fd != our_fd.get()) << "This should always return a fresh fd!";
-      close(fd);
-    }
-    return -ESRCH;
-  }
-  ASSERT(t, child_syscall_result >= 0) << "Failed to recvmsg() in tracee; err="
-                                       << errno_name(-child_syscall_result);
+  // If the child died before reading the message from the socket,
+  // the message will still be in the socket buffer and will be received
+  // the next time we try to send something to a tracee. That's why
+  // before using tracee_socket_receiver_fd we need to drain up to one message
+  // from it.
+  ASSERT(t, child_syscall_result >= 0 || child_syscall_result == -ESRCH)
+    << "Failed to recvmsg() in tracee; err=" << errno_name(-child_syscall_result);
   return child_syscall_result;
 }
 
@@ -691,12 +780,40 @@ remote_ptr<void> AutoRemoteSyscalls::infallible_mmap_syscall_if_alive(
           : infallible_syscall_ptr_if_alive(syscall_number_for_mmap(arch()), addr,
                                             length, prot, flags, child_fd,
                                             offset_bytes);
-  if (ret && (flags & MAP_FIXED)) {
-    ASSERT(t, addr == ret) << "MAP_FIXED at " << addr << " but got " << ret;
+  if (flags & MAP_FIXED) {
+    if (ret) {
+      ASSERT(t, addr == ret) << "MAP_FIXED at " << addr << " but got " << ret;
+    } else {
+      if (!t->vm()->has_mapping(addr)) {
+        KernelMapping km = t->vm()->read_kernel_mapping(t, addr);
+        if (km.size()) {
+          ASSERT(t, km.start() == addr && km.size() == ceil_page_size(length));
+          // The mapping was created. Pretend this call succeeded.
+          ret = addr;
+        }
+      }
+    }
   }
   return ret;
 }
 
+bool AutoRemoteSyscalls::infallible_munmap_syscall_if_alive(
+    remote_ptr<void> addr, size_t length) {
+  long ret = infallible_syscall_if_alive(syscall_number_for_munmap(arch()),
+                                         addr, length);
+  if (ret) {
+    if (t->vm()->has_mapping(addr)) {
+      KernelMapping km = t->vm()->read_kernel_mapping(t, addr);
+      if (!km.size()) {
+        // The unmap happened but the task must have died before
+        // reporting the status.
+        ret = 0;
+      }
+    }
+  }
+  return !ret;
+}
+
 int64_t AutoRemoteSyscalls::infallible_lseek_syscall(int fd, int64_t offset,
                                                      int whence) {
   switch (arch()) {
@@ -730,9 +847,15 @@ void AutoRemoteSyscalls::check_syscall_result(long ret, int syscallno, bool allo
       extra_msg = " opening " + t->read_c_str(t->regs().arg1());
     } else if (is_openat_syscall(syscallno, arch())) {
       extra_msg = " opening " + t->read_c_str(t->regs().arg2());
+    } else if (is_mremap_syscall(syscallno, arch()) ||
+               is_mmap_syscall(syscallno, arch())) {
+      AddressSpace::print_process_maps(t);
     }
     ASSERT(t, false) << "Syscall " << syscall_name(syscallno, arch())
-                     << " failed with errno " << errno_name(-ret) << extra_msg;
+                     << " failed with errno " << errno_name(-ret) << extra_msg
+                     << " arg1=0x" << hex << t->regs().arg1() << " arg2=0x" << t->regs().arg2()
+                     << " arg3=0x" << hex << t->regs().arg3() << " arg4=0x" << t->regs().arg4()
+                     << " arg5=0x" << hex << t->regs().arg5() << " arg6=0x" << t->regs().arg6();
   }
 }
 
@@ -755,6 +878,9 @@ void AutoRemoteSyscalls::finish_direct_mmap(
    * recording. */
   {
     AutoRestoreMem child_str(*this, backing_file_name.c_str());
+    if (word_size(t->arch()) == 4) {
+      backing_file_open_flags |= RR_LARGEFILE_32;
+    }
     fd = infallible_syscall(syscall_number_for_openat(arch()), -1,
                             child_str.get().as_int(),
                             backing_file_open_flags);
