@@ -108,15 +108,33 @@ struct DeschedEvent {
 };
 
 struct PatchSyscallEvent {
-  PatchSyscallEvent() : patch_after_syscall(false) {}
+  PatchSyscallEvent() : patch_trapping_instruction(false),
+    patch_after_syscall(false), patch_vsyscall(false) {}
+  // If true, this patch is for a trapping instruction, not a real syscall
+  bool patch_trapping_instruction;
   // If true, this patch event comes after a syscall (whereas usually they
   // come before). We assume the trace has put us in the correct place
   // and don't try to execute any code to reach this event.
   bool patch_after_syscall;
-  // It true, this patch is for the caller of a vsyscall entry point
+  // If true, this patch is for the caller of a vsyscall entry point
   bool patch_vsyscall;
 };
 
+/**
+ * Sched events track points at which context switches happen that are not
+ * otherwise associated with an rr event.
+ * Also used to record the point at which a tracee is SIGKILLed, which
+ * may require special handling if in the syscallbuf.
+ */
+struct SchedEvent {
+  SchedEvent(remote_code_ptr in_syscallbuf_syscall_hook)
+    : in_syscallbuf_syscall_hook(in_syscallbuf_syscall_hook) {}
+  // If this SchedEvent represents the tracee being SIGKILLed,
+  // and syscall buffering is enabled, this contains the address
+  // of the 'syscall_hook' function, otherwise zero.
+  remote_code_ptr in_syscallbuf_syscall_hook;
+};
+
 struct SyscallbufFlushEvent {
   SyscallbufFlushEvent() {}
   std::vector<mprotect_record> mprotect_records;
@@ -222,6 +240,10 @@ struct SyscallEvent {
   /** Change the architecture for this event. */
   void set_arch(SupportedArch a) { arch_ = a; }
 
+  bool is_exec() const {
+    return is_execve_syscall(number, arch()) || is_execveat_syscall(number, arch());
+  }
+
   SupportedArch arch_;
   // The original (before scratch is set up) arguments to the
   // syscall passed by the tracee.  These are used to detect
@@ -295,6 +317,15 @@ struct Event {
     return patch;
   }
 
+  SchedEvent& Sched() {
+    DEBUG_ASSERT(is_sched_event());
+    return sched_;
+  }
+  const SchedEvent& Sched() const {
+    DEBUG_ASSERT(is_sched_event());
+    return sched_;
+  }
+
   SyscallbufFlushEvent& SyscallbufFlush() {
     DEBUG_ASSERT(EV_SYSCALLBUF_FLUSH == event_type);
     return syscallbuf_flush;
@@ -332,6 +363,9 @@ struct Event {
    * Return true if this is one of the indicated type of events.
    */
   bool is_signal_event() const;
+  bool is_sched_event() const {
+    return event_type == EV_SCHED;
+  }
   bool is_syscall_event() const;
 
   /** Return a string describing this. */
@@ -356,9 +390,14 @@ struct Event {
     auto ev = Event(EV_PATCH_SYSCALL);
     ev.PatchSyscall().patch_after_syscall = false;
     ev.PatchSyscall().patch_vsyscall = false;
+    ev.PatchSyscall().patch_trapping_instruction = false;
+    return ev;
+  }
+  static Event sched() {
+    auto ev = Event(EV_SCHED);
+    ev.Sched().in_syscallbuf_syscall_hook = remote_code_ptr();
     return ev;
   }
-  static Event sched() { return Event(EV_SCHED); }
   static Event seccomp_trap() { return Event(EV_SECCOMP_TRAP); }
   static Event syscallbuf_abort_commit() {
     return Event(EV_SYSCALLBUF_ABORT_COMMIT);
@@ -375,6 +414,7 @@ private:
   union {
     DeschedEvent desched;
     PatchSyscallEvent patch;
+    SchedEvent sched_;
     SignalEvent signal;
     SyscallEvent syscall;
     SyscallbufFlushEvent syscallbuf_flush;
