@@ -4,6 +4,7 @@
 
 #include <string.h>
 
+#include "ReplayTask.h"
 #include "core.h"
 #include "log.h"
 #include "util.h"
@@ -68,6 +69,8 @@ static bool reg_in_range(GdbRegister regno, GdbRegister low, GdbRegister high,
 static const int AVX_FEATURE_BIT = 2;
 static const int PKRU_FEATURE_BIT = 9;
 
+static const uint64_t PKRU_FEATURE_MASK = 1 << PKRU_FEATURE_BIT;
+
 static const size_t xsave_header_offset = 512;
 static const size_t xsave_header_size = 64;
 static const size_t xsave_header_end = xsave_header_offset + xsave_header_size;
@@ -376,9 +379,8 @@ void ExtraRegisters::validate(Task* t) {
   }
 }
 
-static void print_reg(const ExtraRegisters& r, GdbRegister low, GdbRegister hi,
-                      const char* name, FILE* f) {
-  uint8_t buf[128];
+static size_t get_full_value(const ExtraRegisters& r, GdbRegister low, GdbRegister hi,
+                             uint8_t buf[128]) {
   bool defined = false;
   size_t len = r.read_register(buf, low, &defined);
   DEBUG_ASSERT(defined && len <= 64);
@@ -389,9 +391,15 @@ static void print_reg(const ExtraRegisters& r, GdbRegister low, GdbRegister hi,
       len += len2;
     }
   }
-  char out[257];
+  return len;
+}
+
+static string reg_to_string(const ExtraRegisters& r, GdbRegister low, GdbRegister hi) {
+  uint8_t buf[128];
+  size_t len = get_full_value(r, low, hi, buf);
   bool printed_digit = false;
-  char* p = out;
+  char out_buf[257];
+  char* p = out_buf;
   for (int i = len - 1; i >= 0; --i) {
     if (!printed_digit && !buf[i] && i > 0) {
       continue;
@@ -399,7 +407,13 @@ static void print_reg(const ExtraRegisters& r, GdbRegister low, GdbRegister hi,
     p += sprintf(p, printed_digit ? "%02x" : "%x", buf[i]);
     printed_digit = true;
   }
-  fprintf(f, "%s:0x%s", name, out);
+  return out_buf;
+}
+
+static void print_reg(const ExtraRegisters& r, GdbRegister low, GdbRegister hi,
+                      const char* name, FILE* f) {
+  string out = reg_to_string(r, low, hi);
+  fprintf(f, "%s:0x%s", name, out.c_str());
 }
 
 static void print_regs(const ExtraRegisters& r, GdbRegister low, GdbRegister hi,
@@ -575,7 +589,7 @@ bool ExtraRegisters::set_to_raw_data(SupportedArch a, Format format,
      * feature bit. So ignore the PKRU bit here and leave users on their
      * own with respect to RDPKRU.
      */
-    features &= ~(uint64_t(1) << PKRU_FEATURE_BIT);
+    features &= ~PKRU_FEATURE_MASK;
     if (features & ~native_layout.supported_feature_bits) {
       LOG(error) << "Unsupported CPU features found: got " << HEX(features)
                  << " (" << xsave_feature_string(features)
@@ -604,13 +618,8 @@ bool ExtraRegisters::set_to_raw_data(SupportedArch a, Format format,
 
   uint64_t features = features_used(data);
   // OK, now both our native layout and the input layout are using the full
-  // XSAVE header. Copy the header. Make sure to use our updated `features`.
-  memcpy(data_.data() + xsave_header_offset, &features, sizeof(features));
-  memcpy(data_.data() + xsave_header_offset + sizeof(features),
-         data + xsave_header_offset + sizeof(features),
-         xsave_header_size - sizeof(features));
-
-  // Now copy each optional and present area into the right place in our struct
+  // XSAVE header. Copy each optional and present area into the right place
+  // in our struct.
   for (size_t i = 2; i < 64; ++i) {
     if (features & (uint64_t(1) << i)) {
       if (i >= layout.feature_layouts.size()) {
@@ -624,8 +633,27 @@ bool ExtraRegisters::set_to_raw_data(SupportedArch a, Format format,
                    << feature.size << " > " << layout.full_size;
         return false;
       }
+      if (i >= native_layout.feature_layouts.size()) {
+        if (i == PKRU_FEATURE_BIT) {
+          // The native arch doesn't support PKRU.
+          // This must be during replay, and as the comments above explain,
+          // it's OK to not set PKRU during replay on a pre-PKRU CPU, so
+          // we can just ignore this.
+          features &= ~PKRU_FEATURE_MASK;
+          continue;
+        } else {
+          LOG(error) << "Invalid feature " << i << " beyond max layout "
+                     << layout.feature_layouts.size();
+          return false;
+        }
+      }
       const XSaveFeatureLayout& native_feature =
           native_layout.feature_layouts[i];
+      if (native_feature.size == 0 && i == PKRU_FEATURE_BIT) {
+        // See the above comment about PKRU.
+        features &= ~PKRU_FEATURE_MASK;
+        continue;
+      }
       if (feature.size != native_feature.size) {
         LOG(error) << "Feature " << i << " has wrong size " << feature.size
                    << ", expected " << native_feature.size;
@@ -640,6 +668,12 @@ bool ExtraRegisters::set_to_raw_data(SupportedArch a, Format format,
     }
   }
 
+  // Copy the header. Make sure to use our updated `features`.
+  memcpy(data_.data() + xsave_header_offset, &features, sizeof(features));
+  memcpy(data_.data() + xsave_header_offset + sizeof(features),
+         data + xsave_header_offset + sizeof(features),
+         xsave_header_size - sizeof(features));
+
   return true;
 }
 
@@ -745,17 +779,16 @@ void ExtraRegisters::reset() {
       * set in order to get the kernel to properly update the PKRU register
       * value. If this is not set, it has been observed that the PKRU register
       * may occasionally contain "stale" values, particularly after involuntary
-      * context swtiches.
+      * context switches.
       * Avoid this issue by setting the bit if the feature is supported by the
       * CPU.
       */
-      uint64_t pkru_bit = uint64_t(1) << PKRU_FEATURE_BIT;
-      if (xcr0() & pkru_bit) {
+      if (xcr0() & PKRU_FEATURE_MASK) {
         RegData d = xsave_register_data(arch(), arch() == x86_64 ? DREG_64_PKRU : DREG_PKRU);
         DEBUG_ASSERT(d.xsave_feature_bit == PKRU_FEATURE_BIT);
         DEBUG_ASSERT(d.offset + d.size <= (int)data_.size());
         *reinterpret_cast<int*>(data_.data() + d.offset) = 0x55555554;
-        xinuse |= pkru_bit;
+        xinuse |= PKRU_FEATURE_MASK;
       }
 
       memcpy(data_.data() + xinuse_offset, &xinuse, sizeof(xinuse));
@@ -767,4 +800,65 @@ void ExtraRegisters::reset() {
   }
 }
 
+static void compare_regs(const ExtraRegisters& reg1,
+                         const ExtraRegisters& reg2,
+                         GdbRegister low, GdbRegister hi,
+                         int num_regs, const char* name_base,
+                         Registers::Comparison& result) {
+  for (int i = 0; i < num_regs; ++i) {
+    GdbRegister this_low = (GdbRegister)(low + i);
+    GdbRegister this_hi = hi == GdbRegister(0) ? hi : (GdbRegister)(hi + i);
+    uint8_t buf1[128];
+    size_t len1 = get_full_value(reg1, this_low, this_hi, buf1);
+    uint8_t buf2[128];
+    size_t len2 = get_full_value(reg2, this_low, this_hi, buf2);
+    DEBUG_ASSERT(len1 == len2);
+
+    if (!memcmp(buf1, buf2, len1)) {
+      continue;
+    }
+
+    ++result.mismatch_count;
+    if (result.store_mismatches) {
+      char regname[80];
+      sprintf(regname, "%s%d", name_base, i);
+      result.mismatches.push_back({regname, reg_to_string(reg1, this_low, this_hi),
+          reg_to_string(reg2, this_low, this_hi)});
+    }
+  }
+}
+
+void ExtraRegisters::compare_internal(const ExtraRegisters& reg2,
+  Registers::Comparison& result) const {
+  if (arch() != reg2.arch()) {
+    FATAL() << "Can't compare register files with different archs";
+  }
+
+  if (format() == NONE || reg2.format() == NONE) {
+    // Not enough data to check anything
+    return;
+  }
+  if (format() != reg2.format()) {
+    FATAL() << "Can't compare register files with different formats";
+  }
+
+  switch (arch()) {
+    case x86:
+      compare_regs(*this, reg2, DREG_ST0, GdbRegister(0), 8, "st", result);
+      compare_regs(*this, reg2, DREG_XMM0, DREG_YMM0H, 8, "ymm", result);
+      break;
+    case x86_64:
+      compare_regs(*this, reg2, DREG_64_ST0, GdbRegister(0), 8, "st", result);
+      compare_regs(*this, reg2, DREG_64_XMM0, DREG_64_YMM0H, 8, "ymm", result);
+      break;
+    case aarch64:
+      DEBUG_ASSERT(format_ == NT_FPR);
+      compare_regs(*this, reg2, DREG_V0, GdbRegister(0), 32, "v", result);
+      break;
+    default:
+      DEBUG_ASSERT(0 && "Unknown arch");
+      break;
+  }
+}
+
 } // namespace rr
