@@ -9,10 +9,12 @@
 #include <vector>
 
 #include "GdbRegister.h"
+#include "Registers.h"
 #include "kernel_abi.h"
 
 namespace rr {
 
+class ReplayTask;
 struct XSaveLayout;
 
 /**
@@ -142,9 +144,35 @@ public:
 
   void validate(Task* t);
 
+  /**
+   * Return true if |reg1| matches |reg2|.  Passing EXPECT_MISMATCHES
+   * indicates that the caller is using this as a general register
+   * compare and nothing special should be done if the register files
+   * mismatch.  Passing LOG_MISMATCHES will log the registers that don't
+   * match.  Passing BAIL_ON_MISMATCH will additionally abort on
+   * mismatch.
+   * This is conservative; we only return false if we have enough
+   * information to verify that the registers definitely don't match.
+   * The register files must have the same arch.
+   */
+  Registers::Comparison compare_with(const ExtraRegisters& reg2) const {
+    Registers::Comparison result;
+    compare_internal(reg2, result);
+    return result;
+  }
+
+  bool matches(const ExtraRegisters& reg2) const {
+    Registers::Comparison result;
+    result.store_mismatches = false;
+    compare_internal(reg2, result);
+    return !result.mismatch_count;
+  }
+
 private:
   friend class Task;
 
+  void compare_internal(const ExtraRegisters& reg2, Registers::Comparison& result) const;
+
   Format format_;
   SupportedArch arch_;
   std::vector<uint8_t> data_;
