@@ -416,7 +416,9 @@ static string read_target_desc(const char* file_name) {
   string path = resource_path() + share_path + string(file_name);
   stringstream ss;
   FILE* f = fopen(path.c_str(), "r");
-  DEBUG_ASSERT(f);
+  if (f == NULL) {
+      FATAL() << "Failed to load target description file: " << file_name;
+  }
   while (true) {
     int ch = getc(f);
     if (ch == EOF) {
@@ -951,7 +953,7 @@ bool GdbConnection::process_vpacket(char* payload) {
     }
     req = GdbRequest(DREQ_CONT);
     req.cont().run_direction = RUN_FORWARD;
-    req.cont().actions = move(actions);
+    req.cont().actions = std::move(actions);
     return true;
   }
 
@@ -1272,7 +1274,7 @@ bool GdbConnection::process_packet() {
             parser_assert('\0' == *payload);
             payload[0] = tmp;
           }
-          req.watch().conditions.push_back(move(bytes));
+          req.watch().conditions.push_back(std::move(bytes));
         }
       }
       parser_assert('\0' == *payload);
@@ -1544,7 +1546,7 @@ void GdbConnection::notify_stop(GdbThreadId thread, int sig,
   // for the previously-stopped thread when it obviously intends
   // to be making requests about the stopped thread.
   // To make things even better, gdb expects different behavior
-  // for forward continue/interupt and reverse continue.
+  // for forward continue/interrupt and reverse continue.
   if (req.is_resume_request() && req.cont().run_direction == RUN_BACKWARD) {
     LOG(debug) << "Setting query/resume_thread to ANY after reverse continue";
     query_thread = resume_thread = GdbThreadId::ANY;
