@@ -9,6 +9,7 @@
 #include <memory>
 #include <ostream>
 #include <string>
+#include <unordered_set>
 #include <vector>
 
 #include "GdbRegister.h"
@@ -582,7 +583,8 @@ public:
   enum {
     CPU_X86_64 = 0x1,
     CPU_AVX = 0x2,
-    CPU_AARCH64 = 0x4
+    CPU_AARCH64 = 0x4,
+    CPU_PKU = 0x8
   };
   void set_cpu_features(uint32_t features) { cpu_features_ = features; }
   uint32_t cpu_features() const { return cpu_features_; }
@@ -603,6 +605,8 @@ public:
   bool hwbreak_supported() { return hwbreak_supported_; }
   bool swbreak_supported() { return swbreak_supported_; }
 
+  bool is_pass_signal(int sig);
+
 private:
   /**
    * read() incoming data exactly one time, successfully.  May block.
@@ -687,6 +691,10 @@ private:
   // true when "no-ack mode" enabled, in which we don't have
   // to send ack packets back to gdb.  This is a huge perf win.
   bool no_ack;
+  // contains signals (gdb not native) which should be passed directly to the
+  // debuggee without gdb being informed, speeding up
+  // reverse execution
+  std::unordered_set<int> pass_signals;
   ScopedFd sock_fd;
   std::vector<uint8_t> inbuf;  /* buffered input from gdb */
   size_t packetend;            /* index of '#' character */
