@@ -85,7 +85,9 @@ enum CpuMicroarch {
   IntelTigerlake,
   IntelRocketlake,
   IntelAlderlake,
-  LastIntel = IntelAlderlake,
+  IntelRaptorlake,
+  IntelSapphireRapid,
+  LastIntel = IntelSapphireRapid,
   FirstAMD,
   AMDF15R30 = FirstAMD,
   AMDZen,
@@ -103,7 +105,9 @@ enum CpuMicroarch {
   ARMCortexX1,
   AppleM1Icestorm,
   AppleM1Firestorm,
-  LastARM = AppleM1Firestorm,
+  AppleM2Blizzard,
+  AppleM2Avalanche,
+  LastARM = AppleM2Avalanche,
 };
 
 /*
@@ -154,6 +158,8 @@ struct PmuConfig {
 // See Intel 64 and IA32 Architectures Performance Monitoring Events.
 // See check_events from libpfm4.
 static const PmuConfig pmu_configs[] = {
+  { IntelSapphireRapid, "Intel SapphireRapid", 0x5111c4, 0, 0, 125, PMU_TICKS_RCB },
+  { IntelRaptorlake, "Intel Raptorlake", 0x5111c4, 0, 0, 125, PMU_TICKS_RCB },
   { IntelAlderlake, "Intel Alderlake", 0x5111c4, 0, 0, 125, PMU_TICKS_RCB },
   { IntelRocketlake, "Intel Rocketlake", 0x5111c4, 0, 0, 100, PMU_TICKS_RCB },
   { IntelTigerlake, "Intel Tigerlake", 0x5111c4, 0, 0, 100, PMU_TICKS_RCB },
@@ -212,6 +218,10 @@ static const PmuConfig pmu_configs[] = {
     "apple_icestorm_pmu", 0x8c, -1, -1 },
   { AppleM1Firestorm, "Apple M1 Firestorm", 0x90, 0, 0, 1000, PMU_TICKS_TAKEN_BRANCHES,
     "apple_firestorm_pmu", 0x8c, -1, -1 },
+  { AppleM2Blizzard, "Apple M2 Blizzard", 0x90, 0, 0, 1000, PMU_TICKS_TAKEN_BRANCHES,
+    "apple_blizzard_pmu", 0x8c, -1, -1 },
+  { AppleM2Avalanche, "Apple M2 Avalanche", 0x90, 0, 0, 1000, PMU_TICKS_TAKEN_BRANCHES,
+    "apple_avalanche_pmu", 0x8c, -1, -1 },
 };
 
 #define RR_SKID_MAX 10000
@@ -598,8 +608,9 @@ uint32_t PerfCounters::skid_size() {
 }
 
 PerfCounters::PerfCounters(pid_t tid, int cpu_binding,
-                           TicksSemantics ticks_semantics)
-    : tid(tid), pmu_index(get_pmu_index(cpu_binding)), ticks_semantics_(ticks_semantics), started(false), counting(false) {
+                           TicksSemantics ticks_semantics, bool enable)
+    : tid(tid), pmu_index(get_pmu_index(cpu_binding)), ticks_semantics_(ticks_semantics),
+      enable(enable), started(false), counting(false) {
   if (!supports_ticks_semantics(ticks_semantics)) {
     FATAL() << "Ticks semantics " << ticks_semantics << " not supported";
   }
@@ -613,6 +624,9 @@ static void make_counter_async(ScopedFd& fd, int signal) {
 }
 
 void PerfCounters::reset(Ticks ticks_period) {
+  if (!enable) {
+    return;
+  }
   DEBUG_ASSERT(ticks_period >= 0);
   check_pmu(pmu_index);
 
@@ -776,7 +790,7 @@ Ticks PerfCounters::read_ticks(Task* t) {
     if (strex_count > 0) {
       LOG(debug) << strex_count << " strex detected";
       if (!Flags::get().force_things) {
-        CLEAN_FATAL()
+        ASSERT(t, false)
             << strex_count
             << " (speculatively) executed strex instructions detected. \n"
                "On aarch64, rr only supports applications making use of LSE\n"
