@@ -48,8 +48,10 @@ static CpuMicroarch compute_cpu_microarch(const CPUID &cpuid) {
     case 0xd40:
       return ARMNeoverseV1;
     case 0xd41:
+    case 0xd4b: // ARM Cortex A78C
       return ARMCortexA78;
     case 0xd44:
+    case 0xd4c: // ARM Cortex X1C
       return ARMCortexX1;
     case 0xd49:
       return ARMNeoverseN2;
@@ -73,10 +75,16 @@ static CpuMicroarch compute_cpu_microarch(const CPUID &cpuid) {
     switch (cpuid.part) {
     case 0x22:
     case 0x24:
+    case 0x28:
       return AppleM1Icestorm;
     case 0x23:
     case 0x25:
+    case 0x29:
       return AppleM1Firestorm;
+    case 0x32:
+      return AppleM2Blizzard;
+    case 0x33:
+      return AppleM2Avalanche;
     }
     break;
   }
