@@ -51,7 +51,7 @@ struct TraceUuid {
 };
 
 /** Encapsulates additional session state related to recording. */
-class RecordSession : public Session {
+class RecordSession final : public Session {
 public:
   typedef std::shared_ptr<RecordSession> shr_ptr;
 
@@ -73,6 +73,8 @@ public:
       bool force_asan_active = false,
       bool force_tsan_active = false);
 
+  ~RecordSession() override;
+
   const DisableCPUIDFeatures& disable_cpuid_features() const {
     return disable_cpuid_features_;
   }
@@ -175,11 +177,11 @@ public:
   }
 
   virtual Task* new_task(pid_t tid, pid_t rec_tid, uint32_t serial,
-                         SupportedArch a) override;
+                         SupportedArch a, const std::string& name) override;
 
   RecordTask* find_task(pid_t rec_tid) const;
   RecordTask* find_task(const TaskUid& tuid) const;
-  RecordTask* find_detached_proxy_task(pid_t proxt_tid) const;
+  RecordTask* find_detached_proxy_task(pid_t proxy_tid) const;
 
   void on_proxy_detach(RecordTask *t, pid_t new_tid);
 
