@@ -7,11 +7,13 @@
 #include <limits.h>
 #include <linux/perf_event.h>
 #include <sys/prctl.h>
+#include <sys/resource.h>
 #include <sys/syscall.h>
 
 #include "AutoRemoteSyscalls.h"
 #include "PreserveFileMonitor.h"
 #include "RecordSession.h"
+#include "WaitManager.h"
 #include "core.h"
 #include "kernel_abi.h"
 #include "kernel_metadata.h"
@@ -26,7 +28,7 @@ namespace rr {
 
 /**
  * Stores the table of signal dispositions and metadata for an
- * arbitrary set of tasks.  Each of those tasks must own one one of
+ * arbitrary set of tasks.  Each of those tasks must own one of
  * the |refcount|s while they still refer to this.
  */
 struct Sighandler {
@@ -157,6 +159,7 @@ private:
 RecordTask::RecordTask(RecordSession& session, pid_t _tid, uint32_t serial,
                        SupportedArch a)
     : Task(session, _tid, _tid, serial, a),
+      scheduler_token(0),
       ticks_at_last_recorded_syscall_exit(0),
       ip_at_last_recorded_syscall_exit(nullptr),
       time_at_start_of_last_timeslice(0),
@@ -195,7 +198,6 @@ RecordTask::RecordTask(RecordSession& session, pid_t _tid, uint32_t serial,
       next_pmc_interrupt_is_for_user(false),
       did_record_robust_futex_changes(false),
       waiting_for_reap(false),
-      waiting_for_zombie(false),
       waiting_for_ptrace_exit(false),
       retry_syscall_patching(false),
       sent_shutdown_kill(false),
@@ -246,14 +248,19 @@ RecordTask::~RecordTask() {
   }
 
   if (detached_proxy) {
-    // We kept the zombie of the orginal task around to prevent its pid from
+    // Kill it so that proceed_to_exit will reach PTRACE_EVENT_EXIT.
+    ::kill(tid, SIGKILL);
+    // We kept the zombie of the original task around to prevent its pid from
     // being re-used. Reap that now.
     proceed_to_exit();
-    if (!already_reaped() && may_reap()) {
+    if (!was_reaped() && may_reap()) {
       reap();
     }
     did_kill();
   }
+
+  // If this was stopped, notify the scheduler.
+  set_stopped(false);
 }
 
 void RecordTask::record_exit_event(WriteChildTid write_child_tid) {
@@ -354,7 +361,7 @@ void RecordTask::post_exec() {
   ev().Syscall().set_arch(arch());
 
   // The signal mask is inherited across execve so we don't need to invalidate.
-  Task::post_exec(exe_path(), exe_path());
+  Task::post_exec(exe_path());
   if (emulated_ptracer) {
     ASSERT(this, !(emulated_ptracer->arch() == x86 && arch() == x86_64))
         << "We don't support a 32-bit process tracing a 64-bit process";
@@ -406,6 +413,11 @@ template <typename Arch> static void do_preload_init_arch(RecordTask* t) {
   auto random_seed_ptr REMOTE_PTR_FIELD(params.globals.rptr(), random_seed);
   t->write_mem(random_seed_ptr, random_seed);
   t->record_local(random_seed_ptr, &random_seed);
+
+  auto cpu_binding = t->session().cpu_binding();
+  auto cpu_binding_ptr = REMOTE_PTR_FIELD(params.globals.rptr(), cpu_binding);
+  t->write_mem(cpu_binding_ptr, cpu_binding);
+  t->record_local(cpu_binding_ptr, &cpu_binding);
 }
 
 void RecordTask::push_syscall_event(int syscallno) {
@@ -425,12 +437,37 @@ void RecordTask::at_preload_init() {
  * Avoid using low-numbered file descriptors since that can confuse
  * developers.
  */
-static int find_free_file_descriptor(pid_t for_tid) {
-  int fd = 300 + (for_tid % 500);
+static int find_free_file_descriptor(RecordTask* t) {
+  int fd;
+  int fd_limit;
+  struct rlimit limit;
+  // Try to determine what the maximum fd is. If we can't, just
+  // start the search from 0 and bail out if we hit an absurd
+  // number of fds;
+  if (prlimit(t->tgid(), RLIMIT_NOFILE, nullptr, &limit) < 0) {
+    // We might get EPERM if the tracee has changed UID. If that
+    // happens fall back to being slow. (We can probably go faster
+    // using readdir().)
+    ASSERT(t, errno == EPERM) << "Unknown error in prlimit";
+    fd_limit = 128*1024;
+    fd = 0;
+  } else {
+    fd_limit = limit.rlim_cur;
+    fd = max(300, t->fd_table()->last_free_fd());
+  }
+
+  bool searched_from_start = fd == 0;
   while (true) {
+    if (fd >= fd_limit) {
+      ASSERT(t, !searched_from_start) << "No free fds available";
+      fd = 0;
+      searched_from_start = true;
+    }
+
     char buf[PATH_MAX];
-    sprintf(buf, "/proc/%d/fd/%d", for_tid, fd);
+    sprintf(buf, "/proc/%d/fd/%d", t->tid, fd);
     if (access(buf, F_OK) == -1 && errno == ENOENT) {
+      t->fd_table()->set_last_free_fd(fd);
       return fd;
     }
     ++fd;
@@ -438,6 +475,9 @@ static int find_free_file_descriptor(pid_t for_tid) {
 }
 
 template <typename Arch> void RecordTask::init_buffers_arch() {
+  ASSERT(this, as->syscallbuf_enabled())
+      << "Someone called rrcall_init_buffers with syscallbuf disabled?";
+
   // NB: the tracee can't be interrupted with a signal while
   // we're processing the rrcall, because it's masked off all
   // signals.
@@ -448,26 +488,28 @@ template <typename Arch> void RecordTask::init_buffers_arch() {
   auto args = read_mem(child_args);
 
   args.cloned_file_data_fd = -1;
-  if (as->syscallbuf_enabled()) {
-    args.syscallbuf_size = syscallbuf_size = session().syscall_buffer_size();
-    KernelMapping syscallbuf_km = init_syscall_buffer(remote, nullptr);
-    args.syscallbuf_ptr = syscallbuf_child;
-    if (syscallbuf_child != nullptr) {
-      // This needs to be skipped if we couldn't allocate the buffer
-      // since replaying only reads (and advances) the mmap record
-      // if `args.syscallbuf_ptr != nullptr`.
-      auto record_in_trace = trace_writer().write_mapped_region(
-        this, syscallbuf_km, syscallbuf_km.fake_stat(), syscallbuf_km.fsname(),
-        vector<TraceRemoteFd>(),
-        TraceWriter::RR_BUFFER_MAPPING);
-      ASSERT(this, record_in_trace == TraceWriter::DONT_RECORD_IN_TRACE);
-    } else {
-      // This can fail, e.g. if the tracee died unexpectedly.
-      LOG(debug) << "Syscallbuf initialization failed";
-      args.syscallbuf_size = 0;
-    }
+  args.syscallbuf_size = syscallbuf_size = session().syscall_buffer_size();
+  KernelMapping syscallbuf_km = init_syscall_buffer(remote, nullptr);
+  if (!syscallbuf_km.size()) {
+    // Syscallbuf allocation failed. This should mean the child is dead,
+    // but just in case, return an error.
+    remote.regs().set_syscall_result(-ENOMEM);
+    return;
+  }
+
+  args.syscallbuf_ptr = syscallbuf_child;
+  if (syscallbuf_child != nullptr) {
+    // This needs to be skipped if we couldn't allocate the buffer
+    // since replaying only reads (and advances) the mmap record
+    // if `args.syscallbuf_ptr != nullptr`.
+    auto record_in_trace = trace_writer().write_mapped_region(
+      this, syscallbuf_km, syscallbuf_km.fake_stat(), syscallbuf_km.fsname(),
+      vector<TraceRemoteFd>(),
+      TraceWriter::RR_BUFFER_MAPPING);
+    ASSERT(this, record_in_trace == TraceWriter::DONT_RECORD_IN_TRACE);
   } else {
-    args.syscallbuf_ptr = remote_ptr<void>(nullptr);
+    // This can fail, e.g. if the tracee died unexpectedly.
+    LOG(debug) << "Syscallbuf initialization failed";
     args.syscallbuf_size = 0;
   }
 
@@ -483,7 +525,7 @@ template <typename Arch> void RecordTask::init_buffers_arch() {
       ScopedFd clone_file(cloned_file_data_fname.c_str(), O_RDWR | O_CREAT, 0600);
       int cloned_file_data = remote.infallible_send_fd_if_alive(clone_file);
       if (cloned_file_data >= 0) {
-        int free_fd = find_free_file_descriptor(tid);
+        int free_fd = find_free_file_descriptor(this);
         cloned_file_data_fd_child =
             remote.syscall(syscall_number_for_dup3(arch()), cloned_file_data,
                             free_fd, O_CLOEXEC);
@@ -523,6 +565,7 @@ void RecordTask::on_syscall_exit_arch(int syscallno, const Registers& regs) {
   switch (syscallno) {
     // These syscalls affect the sigmask even if they fail.
     case Arch::epoll_pwait:
+    case Arch::epoll_pwait2:
     case Arch::pselect6:
     case Arch::pselect6_time64:
     case Arch::ppoll:
@@ -595,7 +638,7 @@ bool RecordTask::is_at_syscallstub_exit_breakpoint() {
   return i == syscallstub_exit_breakpoint;
 }
 
-bool RecordTask::will_resume_execution(ResumeRequest, WaitRequest,
+void RecordTask::will_resume_execution(ResumeRequest, WaitRequest,
                                        TicksRequest ticks_request, int sig) {
   // We may execute user code, which could lead to an RDTSC or grow-map
   // operation which unblocks SIGSEGV, and we'll need to know whether to
@@ -626,9 +669,7 @@ bool RecordTask::will_resume_execution(ResumeRequest, WaitRequest,
       // We're injecting a signal, so make sure that signal is unblocked.
       sigset &= ~signal_bit(sig);
     }
-    if (!set_sigmask(sigset)) {
-      return false;
-    }
+    set_sigmask(sigset);
   }
 
   // RESUME_NO_TICKS means that tracee code is not going to run so there's no
@@ -666,7 +707,6 @@ bool RecordTask::will_resume_execution(ResumeRequest, WaitRequest,
       }
     }
   }
-  return true;
 }
 
 vector<remote_code_ptr> RecordTask::syscallbuf_syscall_entry_breakpoints() {
@@ -705,7 +745,8 @@ void RecordTask::did_wait() {
     // state, because we do not allow stashed_signals_blocking_more_signals
     // to hold across syscalls (traced or untraced) that change the signal mask.
     ASSERT(this, !blocked_sigs_dirty);
-    xptrace(PTRACE_SETSIGMASK, remote_ptr<void>(8), &blocked_sigs);
+    // If this fails the tracee is on the exit path and its sigmask is irrelevant.
+    ptrace_if_stopped(PTRACE_SETSIGMASK, remote_ptr<void>(8), &blocked_sigs);
   } else if (syscallbuf_child) {
     // The syscallbuf struct is only 32 bytes currently so read the whole thing
     // at once to avoid multiple calls to read_mem. Even though this shouldn't
@@ -797,24 +838,35 @@ void RecordTask::do_ptrace_exit_stop(WaitStatus exit_status) {
 }
 
 void RecordTask::did_reach_zombie() {
-  waiting_for_zombie = false;
   // Remove from address-space and fds list since we really aren't associated
   // with them anymore (and we can't be used to operate on them)
   as->erase_task(this);
   fds->erase_task(this);
 
-  if (!already_reaped()) {
+  if (!was_reaped()) {
     if (may_reap()) {
       reap();
     } else {
       waiting_for_reap = true;
     }
   }
-  if ((already_reaped() || !waiting_for_reap) && !emulated_stop_pending) {
+  if ((was_reaped() || !waiting_for_reap) && !emulated_stop_pending) {
     delete this;
   }
 }
 
+void RecordTask::set_stopped(bool stopped) {
+  if (is_stopped_ == stopped) {
+    return;
+  }
+  is_stopped_ = stopped;
+  if (stopped) {
+    session().scheduler().stopped_task(this);
+  } else {
+    session().scheduler().started_task(this);
+  }
+}
+
 void RecordTask::send_synthetic_SIGCHLD_if_necessary() {
   RecordTask* wake_task = nullptr;
   bool need_signal = false;
@@ -1206,7 +1258,9 @@ SignalResolvedDisposition RecordTask::sig_resolved_disposition(
 
 void RecordTask::set_siginfo(const siginfo_t& si) {
   pending_siginfo = si;
-  ptrace_if_alive(PTRACE_SETSIGINFO, nullptr, (void*)&si);
+  // If this fails, the tracee is on the exit path and its siginfo
+  // is irrelevant.
+  ptrace_if_stopped(PTRACE_SETSIGINFO, nullptr, (void*)&si);
 }
 
 template <typename Arch>
@@ -1268,26 +1322,27 @@ sig_set_t RecordTask::get_sigmask() {
   return blocked_sigs;
 }
 
-bool RecordTask::unblock_signal(int sig) {
+void RecordTask::unblock_signal(int sig) {
   sig_set_t mask = get_sigmask();
   mask &= ~signal_bit(sig);
-  if (!set_sigmask(mask)) {
-    return false;
-  }
+  set_sigmask(mask);
   invalidate_sigmask();
-  return true;
 }
 
-bool RecordTask::set_sigmask(sig_set_t mask) {
+void RecordTask::set_sigmask(sig_set_t mask) {
+  ASSERT(this, is_stopped_);
   int ret = fallible_ptrace(PTRACE_SETSIGMASK, remote_ptr<void>(8), &mask);
   if (ret < 0) {
     if (errno == EIO) {
       FATAL() << "PTRACE_SETSIGMASK not supported; rr requires Linux kernel >= 3.11";
     }
     if (errno == ESRCH) {
-      // Task most likely died while we at the ptrace stop.
-      detected_unexpected_exit = true;
-      return false;
+      // The task has been unexpectedly killed due to SIGKILL or equivalent.
+      // Just pretend we set the mask; it doesn't matter anymore.
+      // Reporting this to the caller is pointless because callers still need
+      // to handle the case where this function succeeds but the task
+      // is unexpectedly killed immediately afterwards.
+      return;
     }
     ASSERT(this, errno == EINVAL);
   } else {
@@ -1295,7 +1350,6 @@ bool RecordTask::set_sigmask(sig_set_t mask) {
                << "SYSCALLBUF_DESCHED_SIGNAL/TIME_SLICE_SIGNAL) while we "
                << " have a stashed signal";
   }
-  return true;
 }
 
 void RecordTask::set_sig_handler_default(int sig) {
@@ -1593,7 +1647,8 @@ bool RecordTask::at_interrupted_non_restartable_signal_modifying_syscall() const
   auto r = regs();
   // XXXkhuey io_uring_enter (not yet supported) can do this too.
   return r.syscall_result_signed() == -EINTR &&
-    is_epoll_pwait_syscall(r.original_syscallno(), arch());
+    (is_epoll_pwait_syscall(r.original_syscallno(), arch()) ||
+     is_epoll_pwait2_syscall(r.original_syscallno(), arch()));
 }
 
 bool RecordTask::is_arm_desched_event_syscall() {
@@ -1611,7 +1666,6 @@ bool RecordTask::may_be_blocked() const {
          emulated_stop_type != NOT_STOPPED ||
          (EV_SIGNAL_DELIVERY == ev().type() &&
           DISPOSITION_FATAL == ev().Signal().disposition) ||
-         waiting_for_zombie ||
          waiting_for_ptrace_exit;
 }
 
@@ -1650,8 +1704,6 @@ int RecordTask::get_ptrace_eventmsg_seccomp_data() {
 
 void RecordTask::record_local(remote_ptr<void> addr, ssize_t num_bytes,
                               const void* data) {
-  maybe_flush_syscallbuf();
-
   ASSERT(this, num_bytes >= 0);
 
   if (!addr) {
@@ -1671,8 +1723,6 @@ bool RecordTask::record_remote_by_local_map(remote_ptr<void> addr,
 }
 
 void RecordTask::record_remote(remote_ptr<void> addr, ssize_t num_bytes) {
-  maybe_flush_syscallbuf();
-
   ASSERT(this, num_bytes >= 0);
 
   if (!addr) {
@@ -1683,7 +1733,21 @@ void RecordTask::record_remote(remote_ptr<void> addr, ssize_t num_bytes) {
     return;
   }
 
-  auto buf = read_mem(addr.cast<uint8_t>(), num_bytes);
+  bool ok = true;
+  auto buf = read_mem(addr.cast<uint8_t>(), num_bytes, &ok);
+  if (!ok) {
+    // Tracee probably died unexpectely. This should only happen
+    // due to SIGKILL racing with our PTRACE_CONT.
+    if (!vm()->find_other_thread_group(this) &&
+        vm()->range_is_private_mapping(MemoryRange(addr, num_bytes))) {
+      // The recording range is mapped private and no other threadgroup shares the
+      // address space, so the new memory contents should be unobservable, and we can
+      // just not record the data.
+      return;
+    }
+    ASSERT(this, false) << "Should have recorded " << num_bytes << " bytes from "
+                        << addr << ", but failed";
+  }
   trace_writer().write_raw(rec_tid, buf.data(), num_bytes, addr);
 }
 
@@ -1755,8 +1819,6 @@ ssize_t RecordTask::record_remote_fallible(remote_ptr<void> addr,
 
 void RecordTask::record_remote_even_if_null(remote_ptr<void> addr,
                                             ssize_t num_bytes) {
-  maybe_flush_syscallbuf();
-
   DEBUG_ASSERT(num_bytes >= 0);
 
   if (!addr) {
@@ -1793,6 +1855,52 @@ void RecordTask::log_pending_events() const {
   }
 }
 
+template <typename Arch>
+static void maybe_handle_set_robust_list_arch(RecordTask* t) {
+  auto remote_locals = AddressSpace::preload_thread_locals_start()
+    .cast<preload_thread_locals<Arch>>();
+  if (!remote_locals) {
+    return;
+  }
+  auto robust_list_ptr = REMOTE_PTR_FIELD(remote_locals, robust_list);
+  auto robust_list = t->read_mem(robust_list_ptr);
+  if (robust_list.len) {
+    t->set_robust_list(robust_list.head.rptr(), robust_list.len);
+
+    auto robust_list_len_ptr = REMOTE_PTR_FIELD(robust_list_ptr, len);
+    t->write_mem(robust_list_len_ptr, static_cast<uint32_t>(0));
+    robust_list.len = 0;
+    t->record_local(robust_list_len_ptr, &robust_list.len);
+  }
+}
+
+static void maybe_handle_set_robust_list(RecordTask* t) {
+  RR_ARCH_FUNCTION(maybe_handle_set_robust_list_arch, t->arch(), t);
+}
+
+template <typename Arch>
+static void maybe_handle_rseq_arch(RecordTask* t) {
+  auto remote_locals = AddressSpace::preload_thread_locals_start()
+    .cast<preload_thread_locals<Arch>>();
+  if (!remote_locals) {
+    return;
+  }
+  auto rseq_ptr = REMOTE_PTR_FIELD(remote_locals, rseq);
+  auto rseq = t->read_mem(rseq_ptr);
+  if (rseq.len) {
+    t->rseq_state = make_unique<RseqState>(rseq.rseq.rptr(), rseq.sig);
+
+    auto rseq_len_ptr = REMOTE_PTR_FIELD(rseq_ptr, len);
+    t->write_mem(rseq_len_ptr, static_cast<uint32_t>(0));
+    rseq.len = 0;
+    t->record_local(rseq_len_ptr, &rseq.len);
+  }
+}
+
+static void maybe_handle_rseq(RecordTask* t) {
+  RR_ARCH_FUNCTION(maybe_handle_rseq_arch, t->arch(), t);
+}
+
 void RecordTask::maybe_flush_syscallbuf() {
   if (EV_SYSCALLBUF_FLUSH == ev().type()) {
     // Already flushing.
@@ -1831,16 +1939,19 @@ void RecordTask::maybe_flush_syscallbuf() {
 
   // Write the entire buffer in one shot without parsing it,
   // because replay will take care of that.
-  if (is_running()) {
+  if (is_stopped()) {
+    record_remote(syscallbuf_child, syscallbuf_data_size());
+  } else {
     vector<uint8_t> buf;
     buf.resize(sizeof(hdr) + hdr.num_rec_bytes);
     memcpy(buf.data(), &hdr, sizeof(hdr));
     read_bytes_helper(syscallbuf_child + 1, hdr.num_rec_bytes,
                       buf.data() + sizeof(hdr));
     record_local(syscallbuf_child, buf.size(), buf.data());
-  } else {
-    record_remote(syscallbuf_child, syscallbuf_data_size());
   }
+  maybe_handle_rseq(this);
+  maybe_handle_set_robust_list(this);
+
   record_current_event();
   pop_event(EV_SYSCALLBUF_FLUSH);
 
@@ -1867,7 +1978,7 @@ void RecordTask::maybe_reset_syscallbuf() {
   }
 }
 
-void RecordTask::record_event(const Event& ev, FlushSyscallbuf flush,
+void RecordTask::record_event(Event ev, FlushSyscallbuf flush,
                               AllowSyscallbufReset reset,
                               const Registers* registers) {
   if (flush == FLUSH_SYSCALLBUF) {
@@ -1981,14 +2092,16 @@ pid_t RecordTask::find_newborn_thread() {
   ASSERT(this, session().is_recording());
   ASSERT(this, ptrace_event() == PTRACE_EVENT_CLONE);
 
-  pid_t hint = get_ptrace_eventmsg<pid_t>();
+  pid_t hint = get_ptrace_eventmsg_pid();
   char path[PATH_MAX];
-  sprintf(path, "/proc/%d/task/%d", tid, hint);
-  struct stat stat_buf;
-  // This should always succeed, but may fail in old kernels due to
-  // a kernel bug. See RecordSession::handle_ptrace_event.
-  if (!session().find_task(hint) && 0 == stat(path, &stat_buf)) {
-    return hint;
+  if (hint >= 0) {
+    sprintf(path, "/proc/%d/task/%d", tid, hint);
+    struct stat stat_buf;
+    // This should always succeed, but may fail in old kernels due to
+    // a kernel bug. See RecordSession::handle_ptrace_event.
+    if (!session().find_task(hint) && 0 == stat(path, &stat_buf)) {
+      return hint;
+    }
   }
 
   sprintf(path, "/proc/%d/task", tid);
@@ -2013,10 +2126,10 @@ pid_t RecordTask::find_newborn_process(pid_t child_parent) {
              ptrace_event() == PTRACE_EVENT_VFORK ||
              ptrace_event() == PTRACE_EVENT_FORK);
 
-  pid_t hint = get_ptrace_eventmsg<pid_t>();
+  pid_t hint = get_ptrace_eventmsg_pid();
   // This should always succeed, but may fail in old kernels due to
   // a kernel bug. See RecordSession::handle_ptrace_event.
-  if (!session().find_task(hint) && get_ppid(hint) == child_parent) {
+  if (hint >= 0 && !session().find_task(hint) && get_ppid(hint) == child_parent) {
     return hint;
   }
 
@@ -2050,7 +2163,7 @@ void RecordTask::update_own_namespace_tid() {
 }
 
 void RecordTask::kill_if_alive() {
-  if (!is_dying()) {
+  if (!seen_ptrace_exit_event()) {
     tgkill(SIGKILL);
   }
 }
@@ -2085,68 +2198,12 @@ bool RecordTask::may_reap() {
 }
 
 void RecordTask::reap() {
-  ASSERT(this, !was_reaped);
+  ASSERT(this, !was_reaped_);
   LOG(debug) << "Reaping " << tid;
-  siginfo_t info;
-  memset(&info, 0, sizeof(info));
-  int ret = waitid(P_PID, tid, &info, WEXITED | WNOHANG);
-  if (ret != 0) {
-    FATAL() << "Unexpected wait status for tid " << tid;
-  }
-  /* The sid_pid == 0 case here is the same as the case below where we're the
-   * group leader whose pid gets stolen.
-   */
-  DEBUG_ASSERT(info.si_pid == tid ||
-               info.si_pid == 0);
-  was_reaped = true;
-}
-
-bool RecordTask::try_wait() {
-  if (wait_unexpected_exit()) {
-    return true;
-  }
-
-  // Check if there is a status change for us
-  WaitStatus status;
-  siginfo_t info;
-  memset(&info, 0, sizeof(siginfo_t));
-  int ret = waitid(P_PID, tid, &info, WSTOPPED | WNOHANG);
-  ASSERT(this, 0 == ret || (-1 == ret && errno == ECHILD)) <<
-    "waitid(" << tid << ", WSTOPPED | NOHANG) failed with "
-                         << ret;
-  LOG(debug) << "waitid(" << tid << ", NOHANG) returns " << ret;
-  if (ret == 0 && info.si_pid == 0) {
-    return false;
-  }
-  if (ret == 0) {
-    status = WaitStatus(info);
-  } else if (ret == -1) {
-    ASSERT(this, errno == ECHILD);
-    // Either we died/are dying unexpectedly, or we were in exec and changed the tid.
-    // Try to differentiate the two situations by seeing if there is an exit
-    // notification ready for us to de-queue, in which case we synthesize an
-    // exit event (but don't actually reap the task, instead leaving that
-    // for the generic cleanup code).
-    int ret = waitid(P_PID, tid, &info, WEXITED | WNOWAIT | WNOHANG);
-    if (ret == 0) {
-      if (info.si_pid == tid) {
-        LOG(debug) << "Synthesizing PTRACE_EVENT_EXIT for zombie process in try_wait " << tid;
-        status = WaitStatus::for_ptrace_event(PTRACE_EVENT_EXIT);
-      } else {
-        // This can happen when the task is in zap_pid_ns_processes waiting for all tasks
-        // in the pid-namespace to exit. It's not in a signal stop, but it's also not
-        // ready to be reaped yet, yet we're still tracing it. Don't wait on this
-        // task, we should be able to reap it later.
-        ASSERT(this, info.si_pid == 0);
-        return false;
-      }
-    } else {
-      ASSERT(this, ret == -1 && errno == ECHILD) << "waitpid failed with " << ret;
-      return false;
-    }
-  }
-  did_waitpid(status);
-  return true;
+  WaitOptions options(tid);
+  options.block_seconds = 0;
+  WaitManager::wait_exit(options);
+  was_reaped_ = true;
 }
 
 static uint64_t read_pid_ns(const RecordTask* t) {
@@ -2165,7 +2222,8 @@ static uint64_t read_pid_ns(const RecordTask* t) {
 }
 
 bool RecordTask::waiting_for_pid_namespace_tasks_to_exit() const {
-  if (tg->tgid_own_namespace != 1) {
+  // read_pid_ns() will fail if we're reaped
+  if (tg->tgid_own_namespace != 1 || was_reaped()) {
     return false;
   }
   // This might be the last live thread for pid-1 in the pid namespace.
@@ -2202,6 +2260,13 @@ bool RecordTask::waiting_for_pid_namespace_tasks_to_exit() const {
   return false;
 }
 
+// Disable chaos mode memory randomization for 32-bit ASAN builds. There isn't
+// much address space to play with and it gets tricky.
+bool RecordTask::enable_chaos_memory_allocations() const {
+  return session().enable_chaos() &&
+    (session().excluded_ranges().empty() || word_size(arch()) >= 8);
+}
+
 int RecordTask::process_depth() const {
   int depth = 0;
   ThreadGroup* tg = this->tg.get();
