@@ -63,7 +63,7 @@ enum SignalDisposition { SIGNAL_DEFAULT, SIGNAL_IGNORE, SIGNAL_HANDLER };
  * Every Task owned by a RecordSession is a RecordTask. Functionality that
  * only applies during recording belongs here.
  */
-class RecordTask : public Task {
+class RecordTask final : public Task {
 public:
   RecordTask(RecordSession& session, pid_t _tid, uint32_t serial,
              SupportedArch a);
@@ -77,10 +77,12 @@ public:
   virtual void post_wait_clone(Task* cloned_from, int flags) override;
   virtual void on_syscall_exit(int syscallno, SupportedArch arch,
                                const Registers& regs) override;
-  virtual bool will_resume_execution(ResumeRequest, WaitRequest, TicksRequest,
+  virtual void will_resume_execution(ResumeRequest, WaitRequest, TicksRequest,
                                      int /*sig*/) override;
   virtual void did_wait() override;
 
+  bool enable_chaos_memory_allocations() const;
+
   std::vector<remote_code_ptr> syscallbuf_syscall_entry_breakpoints();
   bool is_at_syscallbuf_syscall_entry_breakpoint();
   bool is_at_syscallbuf_final_instruction_breakpoint();
@@ -184,8 +186,12 @@ public:
    */
   bool is_waiting_for(RecordTask* t);
 
-  virtual bool already_exited() const override {
-    return waiting_for_reap || waiting_for_zombie;
+  bool already_exited() const override {
+    return waiting_for_reap;
+  }
+
+  bool is_detached_proxy() const override {
+    return detached_proxy;
   }
 
   /**
@@ -329,7 +335,7 @@ public:
   bool at_may_restart_syscall() const;
   /**
    * Return true iff this is at an execution state where
-   * a syscall that modifes isgnals was interrupted but will not
+   * a syscall that modifies signals was interrupted but will not
    * be automatically restarted.
    **/
   bool at_interrupted_non_restartable_signal_modifying_syscall() const;
@@ -417,6 +423,13 @@ public:
   // exists
   bool record_remote_by_local_map(remote_ptr<void> addr, size_t num_bytes);
 
+  template <typename T>
+  void write_and_record(remote_ptr<T> addr, const T& value, bool* ok = nullptr,
+                        uint32_t flags = 0) {
+    write_mem(addr, value, ok, flags);
+    record_local(addr, &value, 1);
+  }
+
   /**
    * Save tracee data to the trace.  |addr| is the address in
    * the address space of this task.
@@ -499,7 +512,10 @@ public:
      */
     DONT_RESET_SYSCALLBUF
   };
-  void record_event(const Event& ev, FlushSyscallbuf flush = FLUSH_SYSCALLBUF,
+  // Take `ev` by value to avoid bugs where we pass in an event in
+  // `pending_events`, which could lead to dangling references when
+  // flushing the syscallbuf manipulates `pending_events`.
+  void record_event(Event ev, FlushSyscallbuf flush = FLUSH_SYSCALLBUF,
                     AllowSyscallbufReset reset = ALLOW_RESET_SYSCALLBUF,
                     const Registers* registers = nullptr);
 
@@ -519,9 +535,7 @@ public:
   pid_t find_newborn_process(pid_t child_parent);
 
   /**
-   * If the process looks alive, kill it. It is recommended to call try_wait(),
-   * on this task before, to make sure liveness is correctly reflected when
-   * making this decision
+   * If the process looks alive, kill it.
    */
   void kill_if_alive();
 
@@ -564,9 +578,8 @@ public:
   sig_set_t read_sigmask_from_process();
   /**
    * Unblock the signal for the process.
-   * Returns `false` if the process died underneath us.
    */
-  bool unblock_signal(int sig);
+  void unblock_signal(int sig);
   /**
    * Set the signal handler to default for the process.
    */
@@ -586,7 +599,7 @@ public:
   bool is_container_init() const { return tg->tgid_own_namespace == 1; }
 
   /**
-   * Linux requires the invariant that that all members of a thread group
+   * Linux requires the invariant that all members of a thread group
    * are reaped before the thread group leader. This determines whether or
    * not we're allowed to attempt reaping this thread or whether doing so
    * risks deadlock.
@@ -599,12 +612,6 @@ public:
    */
   void reap();
 
-  /**
-   * Return true if the status of this has changed, but don't
-   * block.
-   */
-  bool try_wait();
-
   bool waiting_for_pid_namespace_tasks_to_exit() const;
   int process_depth() const;
 
@@ -617,7 +624,18 @@ public:
    */
   void send_synthetic_SIGCHLD_if_necessary();
 
-  bool set_sigmask(sig_set_t mask);
+  void set_sigmask(sig_set_t mask);
+
+  /**
+   * Update the futex robust list head pointer to |list| (which
+   * is of size |len|).
+   */
+  void set_robust_list(remote_ptr<void> list, size_t len) {
+    robust_futex_list = list;
+    robust_futex_list_len = len;
+  }
+
+  void set_stopped(bool stopped) override;
 
 private:
   /* Retrieve the tid of this task from the tracee and store it */
@@ -637,14 +655,6 @@ private:
    * Call this when SYS_sigaction is finishing with |regs|.
    */
   void update_sigaction(const Registers& regs);
-  /**
-   * Update the futex robust list head pointer to |list| (which
-   * is of size |len|).
-   */
-  void set_robust_list(remote_ptr<void> list, size_t len) {
-    robust_futex_list = list;
-    robust_futex_list_len = len;
-  }
 
   template <typename Arch> void init_buffers_arch();
   template <typename Arch>
@@ -658,6 +668,7 @@ private:
   virtual bool post_vm_clone(CloneReason reason, int flags, Task* origin) override;
 
 public:
+  uint64_t scheduler_token;
   std::unique_ptr<TaskSyscallStateBase> syscall_state;
   Ticks ticks_at_last_recorded_syscall_exit;
   remote_code_ptr ip_at_last_recorded_syscall_exit;
@@ -775,10 +786,8 @@ public:
   // Our value for ARCH_GET/SET_CPUID (0 -> generate SIGSEGV, 1 -> do CPUID).
   // Only used if session().has_cpuid_faulting().
   int cpuid_mode;
-  // The current stack of events being processed.  (We use a
-  // deque instead of a stack because we need to iterate the
-  // events.)
-  std::deque<Event> pending_events;
+  // The current stack of events being processed.
+  std::vector<Event> pending_events;
   // Stashed signal-delivery state, ready to be delivered at
   // next opportunity.
   std::deque<StashedSignal> stashed_signals;
@@ -798,9 +807,6 @@ public:
   // This task is just waiting to be reaped.
   bool waiting_for_reap;
 
-  // This task is waiting to reach zombie state
-  bool waiting_for_zombie;
-
   // This task is waiting for a ptrace exit event. It should not
   // be manually run.
   bool waiting_for_ptrace_exit;
