@@ -163,6 +163,7 @@ ReplaySession::ReplaySession(const std::string& dir, const Flags& flags)
       ticks_at_start_of_event(0),
       flags_(flags),
       skip_next_execution_event(false),
+      replay_stops_at_first_execve_(flags.replay_stops_at_first_execve),
       trace_start_time(0) {
   if (trace_in.required_forward_compatibility_version() > FORWARD_COMPATIBILITY_VERSION) {
     CLEAN_FATAL()
@@ -172,6 +173,7 @@ ReplaySession::ReplaySession(const std::string& dir, const Flags& flags)
 
   ticks_semantics_ = trace_in.ticks_semantics();
   rrcall_base_ = trace_in.rrcall_base();
+  syscallbuf_fds_disabled_size_ = trace_in.syscallbuf_fds_disabled_size();
 
   if (!flags.redirect_stdio_file.empty()) {
     tracee_output_fd_ = make_shared<ScopedFd>(flags.redirect_stdio_file.c_str(), O_CREAT | O_TRUNC | O_WRONLY, 0600);
@@ -222,6 +224,7 @@ ReplaySession::ReplaySession(const ReplaySession& other)
       flags_(other.flags_),
       fast_forward_status(other.fast_forward_status),
       skip_next_execution_event(other.skip_next_execution_event),
+      replay_stops_at_first_execve_(other.replay_stops_at_first_execve_),
       trace_start_time(other.trace_start_time) {}
 
 ReplaySession::~ReplaySession() {
@@ -292,6 +295,7 @@ DiversionSession::shr_ptr ReplaySession::clone_diversion() {
   session->tracee_socket = tracee_socket;
   session->tracee_socket_fd_number = tracee_socket_fd_number;
   session->rrcall_base_ = rrcall_base_;
+  session->syscallbuf_fds_disabled_size_ = syscallbuf_fds_disabled_size_;
   LOG(debug) << "  deepfork session is " << session.get();
 
   copy_state_to(*session, emufs(), session->emufs());
@@ -301,8 +305,8 @@ DiversionSession::shr_ptr ReplaySession::clone_diversion() {
 }
 
 Task* ReplaySession::new_task(pid_t tid, pid_t rec_tid, uint32_t serial,
-                              SupportedArch a) {
-  return new ReplayTask(*this, tid, rec_tid, serial, a);
+                              SupportedArch a, const std::string& name) {
+  return new ReplayTask(*this, tid, rec_tid, serial, a, name);
 }
 
 /*static*/ ReplaySession::shr_ptr ReplaySession::create(const string& dir,
@@ -488,7 +492,8 @@ Completion ReplaySession::cont_syscall_boundary(
   } else {
     ResumeRequest resume_how =
         constraints.is_singlestep() ? RESUME_SYSEMU_SINGLESTEP : RESUME_SYSEMU;
-    t->resume_execution(resume_how, RESUME_WAIT, ticks_request);
+    bool ok = t->resume_execution(resume_how, RESUME_WAIT_NO_EXIT, ticks_request);
+    ASSERT(t, ok) << "Tracee died unexpectedly";
   }
 
   switch (t->stop_sig()) {
@@ -521,7 +526,8 @@ Completion ReplaySession::cont_syscall_boundary(
       syscall_seccomp_ordering_ = PTRACE_SYSCALL_BEFORE_SECCOMP;
     }
     // Eat the following event, either a seccomp or syscall notification
-    t->resume_execution(RESUME_SYSEMU, RESUME_WAIT, ticks_request);
+    bool ok = t->resume_execution(RESUME_SYSEMU, RESUME_WAIT_NO_EXIT, ticks_request);
+    ASSERT(t, ok) << "Tracee died unexpectedly";
   }
 
   t->apply_syscall_entry_regs();
@@ -553,18 +559,7 @@ static void emulate_syscall_entry(ReplayTask* t, const TraceFrame& frame,
                                   remote_code_ptr syscall_instruction) {
   Registers r = t->regs();
   r.set_ip(syscall_instruction.increment_by_syscall_insn_length(t->arch()));
-  r.set_original_syscallno(r.syscallno());
-  r.set_orig_arg1(r.arg1());
-  /**
-   * The aarch64 kernel has a quirk where if the syscallno is -1 (and only -1),
-   * it will apply the -ENOSYS result before any ptrace entry stop.
-   * On x86, this happens unconditionally for every syscall, but there the
-   * result isn't shared with arg1, and we usually don't care because we have
-   * access to original_syscallno.
-   */
-  if (is_x86ish(t->arch()) || (t->arch() == aarch64 && r.syscallno() == -1)) {
-    r.set_syscall_result(-ENOSYS);
-  }
+  r.emulate_syscall_entry();
   t->set_regs(r);
   t->canonicalize_regs(frame.event().Syscall().arch());
   t->validate_regs();
@@ -722,14 +717,16 @@ Completion ReplaySession::continue_or_step(ReplayTask* t,
                                            TicksRequest tick_request,
                                            ResumeRequest resume_how) {
   if (constraints.command == RUN_SINGLESTEP) {
-    t->resume_execution(RESUME_SINGLESTEP, RESUME_WAIT, tick_request);
+    bool ok = t->resume_execution(RESUME_SINGLESTEP, RESUME_WAIT_NO_EXIT, tick_request);
+    ASSERT(t, ok) << "Tracee died unexpectedly";
     handle_unrecorded_cpuid_fault(t, constraints);
   } else if (constraints.command == RUN_SINGLESTEP_FAST_FORWARD) {
     fast_forward_status |= fast_forward_through_instruction(
         t, RESUME_SINGLESTEP, constraints.stop_before_states);
     handle_unrecorded_cpuid_fault(t, constraints);
   } else {
-    t->resume_execution(resume_how, RESUME_WAIT, tick_request);
+    bool ok = t->resume_execution(resume_how, RESUME_WAIT_NO_EXIT, tick_request);
+    ASSERT(t, ok) << "Tracee died unexpectedly";
     if (t->stop_sig() == 0) {
       auto type = AddressSpace::rr_page_syscall_from_exit_point(t->arch(), t->ip());
       if (type && type->traced == AddressSpace::UNTRACED) {
@@ -781,17 +778,14 @@ static void guard_overshoot(ReplayTask* t, const Registers& target_regs,
       t->move_ip_before_breakpoint();
     }
     if (closest_matching_regs) {
-      LOG(error)
-          << "Replay diverged; target registers at ticks target mismatched: ";
-      Registers::compare_register_files(t, "rep overshoot", t->regs(), "rec",
-                                        *closest_matching_regs, LOG_MISMATCHES);
+      ASSERT(t, false) << "overshot target ticks=" << target_ticks << " by "
+        << -remaining_ticks << "; target registers at ticks target mismatched: "
+        << "replay != rec: " << t->regs().compare_with(*closest_matching_regs);
     } else {
-      LOG(error) << "Replay diverged; target registers mismatched: ";
-      Registers::compare_register_files(t, "rep overshoot", t->regs(), "rec",
-                                        target_regs, LOG_MISMATCHES);
+      ASSERT(t, false) << "overshot target ticks=" << target_ticks << " by "
+        << -remaining_ticks << "; target registers mismatched: "
+        << "replay != rec: " << t->regs().compare_with(target_regs);
     }
-    ASSERT(t, false) << "overshot target ticks=" << target_ticks << " by "
-                     << -remaining_ticks;
   }
 }
 
@@ -813,19 +807,16 @@ static void guard_unexpected_signal(ReplayTask* t) {
 }
 
 static bool is_same_execution_point(ReplayTask* t, const Registers& rec_regs,
+                                    const ExtraRegisters& rec_extra_regs,
                                     Ticks ticks_left,
                                     Registers* mismatched_regs,
                                     const Registers** mismatched_regs_ptr,
                                     bool in_syscallbuf) {
-  MismatchBehavior behavior =
-      IS_LOGGING(debug) ? LOG_MISMATCHES : EXPECT_MISMATCHES;
-
   if (ticks_left != 0) {
-    LOG(debug) << "  not same execution point: " << ticks_left
-               << " ticks left (@" << rec_regs.ip() << ")";
     if (IS_LOGGING(debug)) {
-      Registers::compare_register_files(t, "(rep)", t->regs(), "(rec)",
-                                        rec_regs, LOG_MISMATCHES);
+      LOG(debug) << "  not same execution point: " << ticks_left
+                 << " ticks left (@" << rec_regs.ip() << ")"
+                 << " replay vs rec: " << t->regs().compare_with(rec_regs);
     }
     return false;
   }
@@ -842,13 +833,20 @@ static bool is_same_execution_point(ReplayTask* t, const Registers& rec_regs,
       *mismatched_regs_ptr = mismatched_regs;
       return false;
     }
-  } else if (!Registers::compare_register_files(t, "rep", t->regs(), "rec", rec_regs,
-                                                behavior)) {
-    LOG(debug) << "  not same execution point: regs differ (@" << rec_regs.ip()
-               << ")";
+  } else if (!t->regs().matches(rec_regs)) {
+    if (IS_LOGGING(debug)) {
+      LOG(debug) << "  not same execution point: regs differ (@" << rec_regs.ip()
+                 << ") replay vs rec: " << t->regs().compare_with(rec_regs);
+    }
     *mismatched_regs = t->regs();
     *mismatched_regs_ptr = mismatched_regs;
     return false;
+  } else if (!t->extra_regs().matches(rec_extra_regs)) {
+    if (IS_LOGGING(debug)) {
+      LOG(debug) << "  not same execution point: extra regs differ (@" << rec_regs.ip()
+                 << ") replay vs rec: " << t->extra_regs().compare_with(rec_extra_regs);
+    }
+    return false;
   }
   LOG(debug) << "  same execution point";
   return true;
@@ -871,6 +869,7 @@ Completion ReplaySession::emulate_async_signal(
   bool in_syscallbuf = !in_syscallbuf_syscall_hook.is_null();
 
   const Registers& regs = trace_frame.regs();
+  const ExtraRegisters& extra_regs = trace_frame.extra_regs();
   remote_code_ptr ip = regs.ip();
 
   /* Step 1: advance to the target ticks (minus a slack region) as
@@ -957,7 +956,7 @@ Completion ReplaySession::emulate_async_signal(
      * Determining whether we're at a debugger trap is
      * surprisingly complicated. */
     bool at_target = is_same_execution_point(
-        t, regs, ticks_left, &mismatched_regs, &mismatched_regs_ptr, in_syscallbuf);
+        t, regs, extra_regs, ticks_left, &mismatched_regs, &mismatched_regs_ptr, in_syscallbuf);
     if (pending_SIGTRAP) {
       TrapReasons trap_reasons = t->compute_trap_reasons();
       BreakpointType breakpoint_type =
@@ -1107,7 +1106,7 @@ Completion ReplaySession::emulate_async_signal(
      * advancing execution. So we allow |advance_to| to proceed and actually
      * reach the desired state.
      */
-    if (!is_same_execution_point(t, regs, ticks_left, &mismatched_regs,
+    if (!is_same_execution_point(t, regs, extra_regs, ticks_left, &mismatched_regs,
                                  &mismatched_regs_ptr, in_syscallbuf)) {
       guard_unexpected_signal(t);
     }
@@ -1276,6 +1275,10 @@ void ReplaySession::prepare_syscallbuf_records(ReplayTask* t, Ticks ticks) {
              << " bytes of syscall records";
 }
 
+#define PRELOAD_GLOBALS_FIELD_AFTER_SYSCALLBUF_FDS_DISABLED(t, f) \
+    REMOTE_PTR_FIELD_MINUS_OFFSET(t->preload_globals, f,          \
+      SYSCALLBUF_FDS_DISABLED_SIZE - t->session().syscallbuf_fds_disabled_size())
+
 /**
  * Returns mprotect_record_count
  */
@@ -1284,9 +1287,10 @@ static uint32_t apply_mprotect_records(ReplayTask* t,
   uint32_t final_mprotect_record_count =
       t->read_mem(REMOTE_PTR_FIELD(t->syscallbuf_child, mprotect_record_count));
   if (skip_mprotect_records < final_mprotect_record_count) {
+    auto mprotect_records_ptr =
+        PRELOAD_GLOBALS_FIELD_AFTER_SYSCALLBUF_FDS_DISABLED(t, mprotect_records[0]);
     auto records =
-        t->read_mem(REMOTE_PTR_FIELD(t->preload_globals, mprotect_records[0]) +
-                        skip_mprotect_records,
+        t->read_mem(mprotect_records_ptr + skip_mprotect_records,
                     final_mprotect_record_count - skip_mprotect_records);
     for (size_t i = 0; i < records.size(); ++i) {
       auto& r = records[i];
@@ -1309,6 +1313,36 @@ static uint32_t apply_mprotect_records(ReplayTask* t,
   return final_mprotect_record_count;
 }
 
+static void write_breakpoint_value(ReplayTask *t, uint64_t breakpoint_value, uint32_t flags = 0)
+{
+  if (t->session().has_trace_quirk(TraceReader::UsesGlobalsInReplay)) {
+    t->write_mem(
+      PRELOAD_GLOBALS_FIELD_AFTER_SYSCALLBUF_FDS_DISABLED(t, reserved_legacy_breakpoint_value),
+      breakpoint_value, nullptr, flags);
+  } else {
+    t->write_mem(remote_ptr<uint64_t>(RR_PAGE_BREAKPOINT_VALUE),
+      breakpoint_value, nullptr, flags);
+  }
+}
+
+template <typename Arch>
+static void maybe_handle_rseq_arch(ReplayTask* t) {
+  auto remote_locals = AddressSpace::preload_thread_locals_start()
+    .cast<preload_thread_locals<Arch>>();
+  if (!remote_locals) {
+    return;
+  }
+  auto rseq_ptr = REMOTE_PTR_FIELD(remote_locals, rseq);
+  auto rseq = t->read_mem(rseq_ptr);
+  if (rseq.len) {
+    t->rseq_state = make_unique<RseqState>(rseq.rseq.rptr(), rseq.sig);
+  }
+}
+
+static void maybe_handle_rseq(ReplayTask* t) {
+  RR_ARCH_FUNCTION(maybe_handle_rseq_arch, t->arch(), t);
+}
+
 /**
  * Replay all the syscalls recorded in the interval between |t|'s
  * current execution point and the next non-syscallbuf event (the one
@@ -1340,7 +1374,7 @@ Completion ReplaySession::flush_syscallbuf(ReplayTask* t,
       ASSERT(t, added);
     } else {
       LOG(debug) << "Adding breakpoint";
-      t->write_mem(REMOTE_PTR_FIELD(t->preload_globals, breakpoint_value),
+      write_breakpoint_value(t,
         (uint64_t)current_step.flush.stop_breakpoint_offset);
     }
 
@@ -1354,7 +1388,7 @@ Completion ReplaySession::flush_syscallbuf(ReplayTask* t,
                                  BKPT_INTERNAL);
     } else {
       LOG(debug) << "Removing breakpoint " << t->status();
-      t->write_mem(REMOTE_PTR_FIELD(t->preload_globals, breakpoint_value), (uint64_t)-1);
+      write_breakpoint_value(t, (uint64_t)-1);
     }
 
     // Account for buffered syscalls just completed
@@ -1391,32 +1425,32 @@ Completion ReplaySession::flush_syscallbuf(ReplayTask* t,
       Registers r = t->regs();
       r.set_ip(remote_brkpt_addr);
       t->set_regs(r);
-
-      return COMPLETE;
+    } else {
+      return INCOMPLETE;
+    }
+  } else {
+    if (t->stop_sig() == SIGTRAP) {
+      return INCOMPLETE;
     }
 
-    return INCOMPLETE;
-  }
+    Registers r = t->regs();
+    ASSERT(t, t->stop_sig() == SIGSEGV && r.ip() == t->vm()->do_breakpoint_fault_addr())
+        << "Replay got unexpected signal (or none) " << t->stop_sig()
+        << " ip " << r.ip() << " breakpoint_fault_addr " << t->vm()->do_breakpoint_fault_addr();
+    r.set_ip(r.ip().increment_by_movrm_insn_length(t->arch()));
+    t->set_regs(r);
 
-  if (t->stop_sig() == SIGTRAP) {
-    return INCOMPLETE;
-  }
-
-  Registers r = t->regs();
-  ASSERT(t, t->stop_sig() == SIGSEGV && r.ip() == t->vm()->do_breakpoint_fault_addr())
-      << "Replay got unexpected signal (or none) " << t->stop_sig()
-      << " ip " << r.ip() << " breakpoint_fault_addr " << t->vm()->do_breakpoint_fault_addr();
-  r.set_ip(r.ip().increment_by_movrm_insn_length(t->arch()));
-  t->set_regs(r);
-
-  if (current_step.flush.recorded_ticks <= t->tick_count() && has_trace_quirk(TraceReader::BufferedSyscallForcedTick)) {
-    // We've reached the breakpoint and executed at least as many ticks as were recorded for the FLUSH_SYSCALLBUF.
-    // That means the flush was actually performed before we left the syscallbuf code, i.e. due to a SIGKILL
-    // in syscallbuf code. We will have recorded another execution event that triggered a flush, but just ignore it,
-    // since we probably already passed it and this task is about to die anyway.
-    skip_next_execution_event = true;
+    if (current_step.flush.recorded_ticks <= t->tick_count() && has_trace_quirk(TraceReader::BufferedSyscallForcedTick)) {
+      // We've reached the breakpoint and executed at least as many ticks as were recorded for the FLUSH_SYSCALLBUF.
+      // That means the flush was actually performed before we left the syscallbuf code, i.e. due to a SIGKILL
+      // in syscallbuf code. We will have recorded another execution event that triggered a flush, but just ignore it,
+      // since we probably already passed it and this task is about to die anyway.
+      skip_next_execution_event = true;
+    }
   }
 
+  maybe_handle_rseq(t);
+  t->apply_all_data_records_from_trace();
   return COMPLETE;
 }
 
@@ -1624,7 +1658,8 @@ static void end_task(ReplayTask* t) {
   r.set_syscallno(syscall_number_for_exit(t->arch()));
   t->set_regs(r);
   // Enter the syscall.
-  t->resume_execution(RESUME_CONT, RESUME_WAIT, RESUME_NO_TICKS);
+  bool ok = t->resume_execution(RESUME_CONT, RESUME_WAIT, RESUME_NO_TICKS);
+  ASSERT(t, ok) << "Tracee died unexpectedly";
   if (t->session().done_initial_exec()) {
     ASSERT(t, t->ptrace_event() == PTRACE_EVENT_EXIT);
     t->did_handle_ptrace_exit_event();
@@ -1638,7 +1673,7 @@ static void end_task(ReplayTask* t) {
 }
 
 Completion ReplaySession::exit_task(ReplayTask* t) {
-  ASSERT(t, !t->seen_ptrace_exit_event);
+  ASSERT(t, !t->seen_ptrace_exit_event());
   // Apply robust-futex updates captured during recording.
   t->apply_all_data_records_from_trace();
   end_task(t);
@@ -1737,11 +1772,11 @@ ReplayTask* ReplaySession::setup_replay_one_trace_frame(ReplayTask* t) {
       current_step.action = TSTEP_RETIRE;
       break;
     case EV_PATCH_SYSCALL:
-      if (ev.PatchSyscall().patch_trapping_instruction ||
-          ev.PatchSyscall().patch_vsyscall) {
-        current_step.action = TSTEP_PATCH_IP;
-      } else if (ev.PatchSyscall().patch_after_syscall) {
+      if (ev.PatchSyscall().patch_after_syscall) {
         current_step.action = TSTEP_PATCH_AFTER_SYSCALL;
+      } else if (ev.PatchSyscall().patch_trapping_instruction ||
+                 ev.PatchSyscall().patch_vsyscall) {
+        current_step.action = TSTEP_PATCH_IP;
       } else {
         current_step.action = TSTEP_PATCH_SYSCALL;
       }
@@ -2051,14 +2086,16 @@ void ReplaySession::detach_tasks(pid_t new_ptracer, ScopedFd& new_tracee_socket_
   for (auto& entry : task_map) {
     Task* t = entry.second;
     t->flush_regs();
-    t->xptrace(PTRACE_DETACH, nullptr, (void*)SIGSTOP);
+    errno = 0;
+    t->fallible_ptrace(PTRACE_DETACH, nullptr, (void*)SIGSTOP);
+    ASSERT(t, !errno) << "failed to detach, with errno " << errno;
   }
   forget_tasks();
 }
 
 void ReplaySession::reattach_tasks(ScopedFd new_tracee_socket, ScopedFd new_tracee_socket_receiver) {
-  tracee_socket = make_shared<ScopedFd>(move(new_tracee_socket));
-  tracee_socket_receiver = make_shared<ScopedFd>(move(new_tracee_socket_receiver));
+  tracee_socket = make_shared<ScopedFd>(std::move(new_tracee_socket));
+  tracee_socket_receiver = make_shared<ScopedFd>(std::move(new_tracee_socket_receiver));
   // Seize all tasks.
   for (auto& entry : task_map) {
     Task* t = entry.second;
@@ -2068,7 +2105,9 @@ void ReplaySession::reattach_tasks(ScopedFd new_tracee_socket, ScopedFd new_trac
   // Get stop events for all tasks
   for (auto& entry : task_map) {
     Task* t = entry.second;
-    t->wait();
+    if (!t->wait()) {
+      FATAL() << "Task " << t->tid << " killed unexpectedly";
+    }
     if (SIGSTOP != t->status().group_stop()) {
       WaitStatus failed_status = t->status();
       FATAL() << "Unexpected stop " << failed_status << " for " << t->tid;
