@@ -3,6 +3,7 @@
 #ifndef RR_REPLAY_TASK_H_
 #define RR_REPLAY_TASK_H_
 
+#include "AutoRemoteSyscalls.h"
 #include "Task.h"
 
 namespace rr {
@@ -13,10 +14,11 @@ class TraceFrame;
  * Every Task owned by a ReplaySession is a ReplayTask. Functionality that
  * only applies during replay belongs here.
  */
-class ReplayTask : public Task {
+class ReplayTask final : public Task {
 public:
   ReplayTask(ReplaySession& session, pid_t _tid, pid_t _rec_tid,
-             uint32_t serial, SupportedArch a);
+             uint32_t serial, SupportedArch a,
+             const std::string& name);
 
   ReplaySession& session() const;
   TraceReader& trace_reader() const;
@@ -26,10 +28,9 @@ public:
    * RRCALL_init_syscall_buffer.  This task must be at the point
    * of *exit from* the rrcall.  Registers will be updated with
    * the return value from the rrcall, which is also returned
-   * from this call.  |map_hint| suggests where to map the
-   * region; see |init_syscallbuf_buffer()|.
+   * from this call..
    */
-  void init_buffers(remote_ptr<void> map_hint);
+  void init_buffers();
   /**
    * Call this method when the exec has completed.
    * `replay_exe` is the name of the real executable file in the trace if we have one,
@@ -39,6 +40,10 @@ public:
    */
   void post_exec_syscall(const std::string& replay_exe, const std::string& original_replay_exe);
 
+  void set_name(AutoRemoteSyscalls& remote, const std::string& name) override;
+
+  void did_prctl_set_prname(remote_ptr<void> child_addr) override;
+
   enum {
     /* The x86 linux 3.5.0-36 kernel packaged with Ubuntu
      * 12.04 has been observed to mutate $esi across
@@ -75,10 +80,16 @@ public:
     seen_sched_in_syscallbuf_syscall_hook = true;
   }
 
-private:
-  template <typename Arch> void init_buffers_arch(remote_ptr<void> map_hint);
+  std::string name() const override {
+    return name_;
+  }
 
-  virtual bool post_vm_clone(CloneReason reason, int flags, Task* origin);
+private:
+  template <typename Arch> void init_buffers_arch();
+
+  bool post_vm_clone(CloneReason reason, int flags, Task* origin) override;
+
+  std::string name_;
 
   // Set to true when we see a sched event with in_syscallbuf_syscall_hook set.
   bool seen_sched_in_syscallbuf_syscall_hook;
