@@ -89,6 +89,7 @@ public:
    * Use ReplayTimeline's breakpoint methods.
    */
   ReplaySession& current_session() { return *current; }
+  const ReplaySession& current_session() const { return *current; }
 
   /**
    * Return a mark for the current state. A checkpoint need not be retained,
@@ -208,14 +209,14 @@ public:
    * replay_step_forward only does one replay step. That means we'll only
    * execute code in current_session().current_task().
    */
-  ReplayResult replay_step_forward(RunCommand command, FrameTime stop_at_time);
+  ReplayResult replay_step_forward(RunCommand command);
 
   ReplayResult reverse_continue(
-      const std::function<bool(ReplayTask* t)>& stop_filter,
+      const std::function<bool(ReplayTask* t, const BreakStatus &)>& stop_filter,
       const std::function<bool()>& interrupt_check);
   ReplayResult reverse_singlestep(
       const TaskUid& tuid, Ticks tuid_ticks,
-      const std::function<bool(ReplayTask* t)>& stop_filter,
+      const std::function<bool(ReplayTask* t, const BreakStatus &)>& stop_filter,
       const std::function<bool()>& interrupt_check);
 
   /**
@@ -270,6 +271,7 @@ private:
     MarkKey(FrameTime trace_time, Ticks ticks, ReplayStepKey step_key)
         : trace_time(trace_time), ticks(ticks), step_key(step_key) {}
     MarkKey(const MarkKey& other) = default;
+    MarkKey& operator=(const MarkKey& other) = default;
     FrameTime trace_time;
     Ticks ticks;
     ReplayStepKey step_key;
@@ -423,7 +425,7 @@ private:
                                       const ReplayResult& result);
   ReplayResult reverse_singlestep(
       const Mark& origin, const TaskUid& step_tuid, Ticks step_ticks,
-      const std::function<bool(ReplayTask* t)>& stop_filter,
+      const std::function<bool(ReplayTask* t, const BreakStatus &)>& stop_filter,
       const std::function<bool()>& interrupt_check);
 
   // Reasonably fast since it just relies on checking the mark map.
