@@ -16,6 +16,7 @@
 #include "EmuFs.h"
 #include "Flags.h"
 #include "PerfCounters.h"
+#include "RecordSession.h"
 #include "RecordTask.h"
 #include "Task.h"
 #include "ThreadGroup.h"
@@ -46,6 +47,7 @@ Session::Session()
       tracee_socket_fd_number(0),
       next_task_serial_(1),
       rrcall_base_(RR_CALL_BASE),
+      syscallbuf_fds_disabled_size_(SYSCALLBUF_FDS_DISABLED_SIZE),
       syscall_seccomp_ordering_(PTRACE_SYSCALL_BEFORE_SECCOMP_UNKNOWN),
       ticks_semantics_(PerfCounters::default_ticks_semantics()),
       done_initial_exec_(false),
@@ -67,11 +69,13 @@ Session::Session(const Session& other) {
   next_task_serial_ = other.next_task_serial_;
   done_initial_exec_ = other.done_initial_exec_;
   rrcall_base_ = other.rrcall_base_;
+  syscallbuf_fds_disabled_size_ = other.syscallbuf_fds_disabled_size_;
   visible_execution_ = other.visible_execution_;
   tracee_socket = other.tracee_socket;
   tracee_socket_receiver = other.tracee_socket_receiver;
   tracee_socket_fd_number = other.tracee_socket_fd_number;
   ticks_semantics_ = other.ticks_semantics_;
+  original_affinity_ = other.original_affinity_;
 }
 
 void Session::on_create(ThreadGroup* tg) { thread_group_map_[tg->tguid()] = tg; }
@@ -146,7 +150,7 @@ ThreadGroup::shr_ptr Session::clone(Task* t, ThreadGroup::shr_ptr tg) {
 }
 
 Task* Session::new_task(pid_t tid, pid_t rec_tid, uint32_t serial,
-                        SupportedArch a) {
+                        SupportedArch a, const std::string&) {
   return new Task(*this, tid, rec_tid, serial, a);
 }
 
@@ -439,55 +443,74 @@ static void remap_shared_mmap(AutoRemoteSyscalls& remote, EmuFs& emu_fs,
 /*static*/ const char* Session::rr_mapping_prefix() { return "/rr-shared-"; }
 
 KernelMapping Session::create_shared_mmap(
-    AutoRemoteSyscalls& remote, size_t size, remote_ptr<void> map_hint,
+    AutoRemoteSyscalls& remote, size_t size, remote_ptr<void> required_child_addr,
     const char* name, int tracee_prot, int tracee_flags,
-    MonitoredSharedMemory::shr_ptr&& monitored) {
+    MonitoredSharedMemory::shr_ptr monitored) {
+  Task* t = remote.task();
   static int nonce = 0;
   // Create the segment we'll share with the tracee.
   char path[PATH_MAX];
   snprintf(path, sizeof(path) - 1, "%s%s%s-%d-%d", tmp_dir(),
-           rr_mapping_prefix(), name, remote.task()->real_tgid(), nonce++);
+           rr_mapping_prefix(), name, t->real_tgid(), nonce++);
 
   ScopedFd shmem_fd(path, O_CREAT | O_EXCL | O_RDWR);
-  ASSERT(remote.task(), shmem_fd.is_open());
+  ASSERT(t, shmem_fd.is_open());
   /* Remove the fs name so that we don't have to worry about
    * cleaning up this segment in error conditions. */
   unlink(path);
 
-  KernelMapping km;
-  int child_shmem_fd = remote.infallible_send_fd_if_alive(shmem_fd);
-  if (child_shmem_fd < 0) {
-    return km;
-  }
-  LOG(debug) << "created shmem segment " << path;
-
-  // Map the segment in ours and the tracee's address spaces.
-  void* map_addr;
-  int flags = MAP_SHARED;
-  if (!map_hint.is_null()) {
-    flags |= MAP_FIXED;
-  }
-  remote_ptr<void> child_map_addr = remote.infallible_mmap_syscall_if_alive(
-      map_hint, size, tracee_prot, flags, child_shmem_fd, 0);
-  if (!child_map_addr) {
-    // tracee unexpectedly died
-    return km;
-  }
-
-  if ((void*)-1 == (map_addr = mmap(nullptr, size, PROT_READ | PROT_WRITE,
-                                    MAP_SHARED, shmem_fd, 0))) {
+  void* map_addr = mmap(nullptr, size, PROT_READ | PROT_WRITE,
+                        MAP_SHARED, shmem_fd, 0);
+  if (map_addr == MAP_FAILED) {
     FATAL() << "Failed to mmap shmem region";
   }
   resize_shmem_segment(shmem_fd, size);
 
+  remote_ptr<void> child_map_addr = required_child_addr;
+  if (child_map_addr.is_null()) {
+    if (t->session().is_recording() &&
+        static_cast<RecordTask*>(t)->enable_chaos_memory_allocations()) {
+      child_map_addr = t->vm()->chaos_mode_find_free_memory(static_cast<RecordTask*>(t),
+          size, nullptr);
+    } else {
+      child_map_addr = t->vm()->find_free_memory(t, size, RR_PAGE_ADDR,
+          AddressSpace::FindFreeMemoryPolicy::USE_LAST_FREE_HINT);
+      if (!child_map_addr) {
+        FATAL() << "Can't find free memory for shared mmap";
+      }
+    }
+  }
+
   struct stat st;
-  ASSERT(remote.task(), 0 == ::fstat(shmem_fd, &st));
-  km = remote.task()->vm()->map(
-      remote.task(), child_map_addr, size, tracee_prot, flags | tracee_flags, 0,
+  ASSERT(t, 0 == ::fstat(shmem_fd, &st));
+  int flags = MAP_SHARED;
+  if (!required_child_addr.is_null()) {
+    flags |= MAP_FIXED;
+  }
+
+  int child_shmem_fd = remote.infallible_send_fd_if_alive(shmem_fd);
+  if (child_shmem_fd < 0) {
+    return KernelMapping();
+  }
+  LOG(debug) << "created shmem segment " << path;
+
+  // Map the segment in ours and the tracee's address spaces.
+  remote_ptr<void> addr = remote.infallible_mmap_syscall_if_alive(
+      child_map_addr, size, tracee_prot, flags | MAP_FIXED, child_shmem_fd, 0);
+  if (!addr) {
+    // tracee unexpectedly died.
+    // We leak the fd; cleaning it up is probably impossible/unnecessary.
+    return KernelMapping();
+  }
+
+  // Note the mapping after we successfully created it in the child.
+  // If the child mapping fails for some reason (e.g. SIGKILL) we still
+  // want our cache to be correct (and not contain the mapping).
+  KernelMapping km = t->vm()->map(
+      t, child_map_addr, size, tracee_prot, flags | tracee_flags, 0,
       path, st.st_dev, st.st_ino, nullptr, nullptr, nullptr, map_addr,
       std::move(monitored));
 
-  shmem_fd.close();
   remote.infallible_close_syscall_if_alive(child_shmem_fd);
   return km;
 }
@@ -526,7 +549,7 @@ static char* extract_name(char* name_buffer, size_t buffer_size) {
 
 const AddressSpace::Mapping Session::recreate_shared_mmap(
     AutoRemoteSyscalls& remote, const AddressSpace::Mapping& m,
-    PreserveContents preserve, MonitoredSharedMemory::shr_ptr&& monitored) {
+    PreserveContents preserve, MonitoredSharedMemory::shr_ptr monitored) {
   char name[PATH_MAX];
   strncpy(name, m.map.fsname().c_str(), sizeof(name) - 1);
   name[sizeof(name) - 1] = 0;
@@ -554,13 +577,13 @@ const AddressSpace::Mapping Session::recreate_shared_mmap(
   return new_map;
 }
 
-const AddressSpace::Mapping& Session::steal_mapping(
+AddressSpace::Mapping Session::steal_mapping(
     AutoRemoteSyscalls& remote, const AddressSpace::Mapping& m,
-    MonitoredSharedMemory::shr_ptr&& monitored) {
+    MonitoredSharedMemory::shr_ptr monitored) {
   // We will include the name of the full path of the original mapping in the
   // name of the shared mapping, replacing slashes by dashes.
   char name[PATH_MAX - 40];
-  strncpy(name, m.map.fsname().c_str(), sizeof(name));
+  strncpy(name, m.map.fsname().c_str(), sizeof(name)-1);
   name[sizeof(name) - 1] = '\0';
   for (char* ptr = name; *ptr != '\0'; ++ptr) {
     if (*ptr == '/') {
@@ -589,16 +612,17 @@ void Session::make_private_shared(AutoRemoteSyscalls& remote,
   // Find a place to map the current segment to temporarily
   remote_ptr<void> start = m.map.start();
   size_t sz = m.map.size();
-  remote_ptr<void> free_mem = remote.task()->vm()->find_free_memory(sz);
+  remote_ptr<void> free_mem = remote.task()->vm()->find_free_memory(remote.task(), sz);
   remote.infallible_syscall(syscall_number_for_mremap(remote.arch()), start, sz,
                             sz, MREMAP_MAYMOVE | MREMAP_FIXED, free_mem);
-  remote.task()->vm()->remap(remote.task(), start, sz, free_mem, sz);
+  remote.task()->vm()->remap(remote.task(), start, sz, free_mem, sz,
+                             MREMAP_MAYMOVE | MREMAP_FIXED);
 
   // AutoRemoteSyscalls may have gotten unlucky and picked the old stack
   // segment as it's scratch space, reevaluate that choice
   AutoRemoteSyscalls remote2(remote.task());
 
-  const AddressSpace::Mapping& new_m = steal_mapping(remote2, m);
+  AddressSpace::Mapping new_m = steal_mapping(remote2, m);
 
   if (!new_m.local_addr) {
     return;
@@ -704,7 +728,7 @@ void Session::copy_state_to(Session& dest, EmuFs& emu_fs, EmuFs& dest_emu_fs) {
 
     group.clone_leader_state = group_leader->capture_state();
   }
-  dest.clone_completion = move(completion);
+  dest.clone_completion = std::move(completion);
 
   DEBUG_ASSERT(dest.vms().size() > 0);
 }
@@ -735,6 +759,8 @@ static bool set_cpu_affinity(int cpu) {
 }
 
 void Session::do_bind_cpu() {
+  sched_getaffinity(0, sizeof(original_affinity_), &original_affinity_);
+
   int cpu_index = this->cpu_binding();
   if (cpu_index >= 0) {
     // Set CPU affinity now, after we've created any helper threads
