@@ -289,9 +289,13 @@ public:
   bool is_replaying() { return as_replay() != nullptr; }
   bool is_diversion() { return as_diversion() != nullptr; }
 
+  // Returns true if execution should be "visible", i.e. it's the main
+  // session of a recording or a replay whose output could be echoed.
   bool visible_execution() const { return visible_execution_; }
   void set_visible_execution(bool visible) { visible_execution_ = visible; }
 
+  virtual bool need_performance_counters() const { return true; }
+
   struct Statistics {
     Statistics()
         : bytes_written(0), ticks_processed(0), syscalls_performed(0) {}
@@ -309,16 +313,19 @@ public:
   Statistics statistics() { return statistics_; }
 
   virtual Task* new_task(pid_t tid, pid_t rec_tid, uint32_t serial,
-                         SupportedArch a);
+                         SupportedArch a, const std::string& name);
 
   std::string read_spawned_task_error() const;
 
-  /* Returns an empty mapping if the tracee died */
+  /* Returns an empty mapping if the tracee died.
+   * If map_address is non-null then we must use that address in the tracee,
+   * otherwise we select the address.
+   */
   static KernelMapping create_shared_mmap(
-      AutoRemoteSyscalls& remote, size_t size, remote_ptr<void> map_hint,
+      AutoRemoteSyscalls& remote, size_t size, remote_ptr<void> required_child_addr,
       const char* name, int tracee_prot = PROT_READ | PROT_WRITE,
       int tracee_flags = 0,
-      MonitoredSharedMemory::shr_ptr&& monitored = nullptr);
+      MonitoredSharedMemory::shr_ptr monitored = nullptr);
 
   static void make_private_shared(AutoRemoteSyscalls& remote,
                                   const AddressSpace::Mapping m);
@@ -335,16 +342,16 @@ public:
   static const AddressSpace::Mapping recreate_shared_mmap(
       AutoRemoteSyscalls& remote, const AddressSpace::Mapping& m,
       PreserveContents preserve = DISCARD_CONTENTS,
-      MonitoredSharedMemory::shr_ptr&& monitored = nullptr);
+      MonitoredSharedMemory::shr_ptr monitored = nullptr);
 
   /* Takes a mapping and replaces it by one that is shared between rr and
      the tracee. The caller is responsible for filling the contents of the
       new mapping.
       Returns an empty mapping if the tracee unexpectedly died.
    */
-  static const AddressSpace::Mapping& steal_mapping(
+  static AddressSpace::Mapping steal_mapping(
       AutoRemoteSyscalls& remote, const AddressSpace::Mapping& m,
-      MonitoredSharedMemory::shr_ptr&& monitored = nullptr);
+      MonitoredSharedMemory::shr_ptr monitored = nullptr);
 
   enum PtraceSyscallBeforeSeccomp {
     PTRACE_SYSCALL_BEFORE_SECCOMP,
@@ -359,6 +366,7 @@ public:
   static const char* rr_mapping_prefix();
 
   ScopedFd& tracee_socket_fd() { return *tracee_socket; }
+  // Before using this, it must be drained. See AutoRemoteSyscalls.
   ScopedFd& tracee_socket_receiver_fd() { return *tracee_socket_receiver; }
   int tracee_fd_number() const { return tracee_socket_fd_number; }
 
@@ -394,11 +402,16 @@ public:
   int syscall_number_for_rrcall_rdtsc() const {
     return SYS_rrcall_rdtsc - RR_CALL_BASE + rrcall_base_;
   }
+  uint32_t syscallbuf_fds_disabled_size() const {
+    return syscallbuf_fds_disabled_size_;
+  }
 
   /* Bind the current process to the a CPU as specified in the session options
      or trace */
   void do_bind_cpu();
 
+  cpu_set_t original_affinity() const { return original_affinity_; }
+
   const ThreadGroupMap& thread_group_map() const { return thread_group_map_; }
 
   virtual int tracee_output_fd(int dflt) {
@@ -447,10 +460,13 @@ protected:
   ScopedFd spawned_task_error_fd_;
 
   int rrcall_base_;
+  uint32_t syscallbuf_fds_disabled_size_;
   PtraceSyscallBeforeSeccomp syscall_seccomp_ordering_;
 
   TicksSemantics ticks_semantics_;
 
+  cpu_set_t original_affinity_;
+
   /**
    * True if we've done an exec so tracees are now in a state that will be
    * consistent across record and replay.
