@@ -26,7 +26,7 @@ class Session;
  * of the thread grouping during record. During replay, we put each task
  * into its own thread group.
  */
-class ThreadGroup : public HasTaskSet {
+class ThreadGroup final : public HasTaskSet {
 public:
   ThreadGroup(Session* session, ThreadGroup* parent,
               pid_t tgid, pid_t thid_own_namespace,
