@@ -25,8 +25,10 @@
 #include "TaskishUid.h"
 #include "core.h"
 #include "kernel_abi.h"
+#include "kernel_metadata.h"
 #include "kernel_supplement.h"
 #include "log.h"
+#include "preload/preload_interface.h"
 #include "rr_trace.capnp.h"
 #include "util.h"
 
@@ -117,7 +119,7 @@ string trace_save_dir() {
   return output_dir ? output_dir : default_rr_trace_dir();
 }
 
-static string latest_trace_symlink() {
+string latest_trace_symlink() {
   return trace_save_dir() + "/latest-trace";
 }
 
@@ -171,9 +173,7 @@ public:
     return maxBytes;
   }
   virtual void skip(size_t bytes) {
-    if (!reader.skip(bytes)) {
-      throw IOException();
-    }
+    reader.skip(bytes);
   }
   virtual kj::ArrayPtr<const capnp::byte> tryGetReadBuffer() {
     const uint8_t* p;
@@ -337,18 +337,15 @@ static void to_trace_signal(trace::Signal::Builder signal, const Event& ev) {
 }
 
 static Event from_trace_signal(EventType type, trace::Signal::Reader signal) {
-  if (signal.getSiginfoArch() != to_trace_arch(NativeArch::arch())) {
-    // XXX if we want to handle consumption of rr traces created on a different
-    // architecture rr build than we're running now, we should convert siginfo
-    // formats here.
-    FATAL() << "Unsupported siginfo arch";
-  }
+  union {
+    NativeArch::siginfo_t native_siginfo;
+    siginfo_t system_siginfo;
+  } si;
   auto siginfo = signal.getSiginfo();
-  if (siginfo.size() != sizeof(siginfo_t)) {
-    FATAL() << "Bad siginfo";
-  }
+  si.native_siginfo = convert_to_native_siginfo(from_trace_arch(signal.getSiginfoArch()),
+      siginfo.begin(), siginfo.size());
   return Event(type,
-               SignalEvent(*reinterpret_cast<const siginfo_t*>(siginfo.begin()),
+               SignalEvent(si.system_siginfo,
                            signal.getDeterministic() ? DETERMINISTIC_SIG
                                                      : NONDETERMINISTIC_SIG,
                            from_trace_disposition(signal.getDisposition())));
@@ -532,14 +529,16 @@ void TraceWriter::write_frame(RecordTask* t, const Event& ev,
   tick_time();
 }
 
-TraceFrame TraceReader::read_frame() {
+TraceFrame TraceReader::read_frame(FrameTime skip_before) {
   auto& events = reader(EVENTS);
   word buf[reasonable_frame_message_words];
   CompressedReaderInputStream stream(events);
   PackedMessageReader frame_msg(stream, ReaderOptions(), buf);
-  trace::Frame::Reader frame = frame_msg.getRoot<trace::Frame>();
-
   tick_time();
+  TraceFrame ret;
+  ret.global_time = time();
+
+  trace::Frame::Reader frame = frame_msg.getRoot<trace::Frame>();
 
   auto mem_writes = frame.getMemWrites();
   raw_recs.resize(mem_writes.size());
@@ -556,8 +555,10 @@ TraceFrame TraceReader::read_frame() {
     raw_recs[i] = { w.getAddr(), (size_t)w.getSize(), i32_to_tid(w.getTid()), h };
   }
 
-  TraceFrame ret;
-  ret.global_time = time();
+  if (ret.global_time < skip_before) {
+    return ret;
+  }
+
   ret.tid_ = i32_to_tid(frame.getTid());
   ret.ticks_ = frame.getTicks();
   if (ret.ticks_ < 0) {
@@ -1252,7 +1253,7 @@ bool TraceReader::read_raw_data_for_frame_with_holes(RawDataWithHoles& d) {
   auto& rec = raw_recs[raw_recs.size() - 1];
   d.rec_tid = rec.rec_tid;
   d.addr = rec.addr;
-  d.holes = move(rec.holes);
+  d.holes = std::move(rec.holes);
   size_t data_size = rec.size;
   for (auto& h : d.holes) {
     data_size -= h.size;
@@ -1410,6 +1411,7 @@ void TraceWriter::close(CloseStatus status, const TraceUuid* uuid) {
   header.setRequiredForwardCompatibilityVersion(FORWARD_COMPATIBILITY_VERSION);
   header.setPreloadThreadLocalsRecorded(true);
   header.setRrcallBase(syscall_number_for_rrcall_init_preload(x86_64));
+  header.setSyscallbufFdsDisabledSize(SYSCALLBUF_FDS_DISABLED_SIZE);
 
   header.setNativeArch(to_trace_arch(NativeArch::arch()));
   if (NativeArch::is_x86ish())
@@ -1431,6 +1433,7 @@ void TraceWriter::close(CloseStatus status, const TraceUuid* uuid) {
     quirks.setSpecialLibrrpage(false);
     quirks.setPkeyAllocRecordedExtraRegs(true);
     quirks.setBufferedSyscallForcedTick(true);
+    quirks.setUsesGlobalsInReplay(false);
   }
   // Add a random UUID to the trace metadata. This lets tools identify a trace
   // easily.
@@ -1594,6 +1597,7 @@ TraceReader::TraceReader(const string& dir)
   preload_thread_locals_recorded_ = header.getPreloadThreadLocalsRecorded();
   ticks_semantics_ = from_trace_ticks_semantics(header.getTicksSemantics());
   rrcall_base_ = header.getRrcallBase();
+  syscallbuf_fds_disabled_size_ = header.getSyscallbufFdsDisabledSize();
   required_forward_compatibility_version_ = header.getRequiredForwardCompatibilityVersion();
   quirks_ = 0;
   {
@@ -1610,6 +1614,9 @@ TraceReader::TraceReader(const string& dir)
     if (quirks.getBufferedSyscallForcedTick()) {
       quirks_ |= BufferedSyscallForcedTick;
     }
+    if (quirks.getUsesGlobalsInReplay()) {
+      quirks_ |= UsesGlobalsInReplay;
+    }
   }
   Data::Reader uuid = header.getUuid();
   uuid_ = unique_ptr<TraceUuid>(new TraceUuid());
