@@ -243,7 +243,7 @@ public:
   bool good() const;
 
   /**
-   * Create a trace where the tracess are bound to cpu |bind_to_cpu|. This
+   * Create a trace where the traces are bound to cpu |bind_to_cpu|. This
    * data is recorded in the trace. If |bind_to_cpu| is -1 then the tracees
    * were not bound.
    * The trace name is determined by |file_name| and _RR_TRACE_DIR (if set)
@@ -353,8 +353,11 @@ public:
    * NB: reading a trace frame has the side effect of ticking
    * the global time to match the time recorded in the trace
    * frame.
+   *
+   * For all frames before `skip_before` we only fill in the `global_time`
+   * field. (Raw data and maps are still accessible.)
    */
-  TraceFrame read_frame();
+  TraceFrame read_frame(FrameTime skip_before = 0);
 
   /**
    * Read the next mapped region descriptor and return it.
@@ -464,6 +467,7 @@ public:
 
   // The base syscall number for rr syscalls in this trace
   int rrcall_base() const { return rrcall_base_; }
+  uint32_t syscallbuf_fds_disabled_size() const { return syscallbuf_fds_disabled_size_; }
 
   SupportedArch arch() const { return arch_; }
 
@@ -484,7 +488,9 @@ public:
     // Whether this trace recorded extra regs for pkey_alloc(2).
     PkeyAllocRecordedExtraRegs = 0x4,
     // Whether this trace forced a tick after buffered syscalls.
-    BufferedSyscallForcedTick = 0x8
+    BufferedSyscallForcedTick = 0x8,
+    // Whether this trace requires globals.is_replay to be toggled.
+    UsesGlobalsInReplay = 0x10,
   };
 
   int quirks() const { return quirks_; }
@@ -509,13 +515,15 @@ private:
   bool chaos_mode_known_;
   bool chaos_mode_;
   int rrcall_base_;
+  uint32_t syscallbuf_fds_disabled_size_;
   int required_forward_compatibility_version_;
   SupportedArch arch_;
   int quirks_;
 };
 
-extern std::string trace_save_dir();
-extern std::string resolve_trace_name(const std::string& trace_name);
+std::string trace_save_dir();
+std::string resolve_trace_name(const std::string& trace_name);
+std::string latest_trace_symlink();
 
 } // namespace rr
 
