@@ -8,14 +8,14 @@
 
 int main(__attribute__((unused)) int argc, char** argv) {
   int bits_match = atoi(argv[1]);
-  char* p1 = mmap(NULL, PAGE_SIZE, PROT_READ | PROT_WRITE,
+  char* p1 = mmap(NULL, get_page_size(), PROT_READ | PROT_WRITE,
                   MAP_ANONYMOUS | MAP_PRIVATE, -1, 0);
-  char* p2 = mmap(NULL, PAGE_SIZE, PROT_READ | PROT_WRITE,
+  char* p2 = mmap(NULL, get_page_size(), PROT_READ | PROT_WRITE,
                   MAP_ANONYMOUS | MAP_PRIVATE, -1, 0);
   intptr_t delta = p2 - p1;
   intptr_t mask = ((1 << bits_match) - 1) << 12;
 
-  test_assert(PAGE_SIZE == (1 << 12));
+  test_assert(get_page_size() == (1 << 12));
   test_assert(p1 != MAP_FAILED);
   test_assert(p2 != MAP_FAILED);
 
