@@ -102,8 +102,7 @@ static bool try_handle_trapped_instruction(RecordTask* t, siginfo_t* si) {
   if (trapped_instruction == TrappedInstruction::RDTSC ||
       trapped_instruction == TrappedInstruction::RDTSCP) {
     if (trapped_instruction == TrappedInstruction::RDTSC &&
-        t->vm()->monkeypatcher().try_patch_trapping_instruction(t,
-         trapped_instruction_len(trapped_instruction))) {
+        t->vm()->monkeypatcher().try_patch_trapping_instruction(t, len, true)) {
       Event ev = Event::patch_syscall();
       ev.PatchSyscall().patch_trapping_instruction = true;
       t->record_event(ev);
@@ -128,8 +127,20 @@ static bool try_handle_trapped_instruction(RecordTask* t, siginfo_t* si) {
 
   r.set_ip(r.ip() + len);
   t->set_regs(r);
+  t->record_event(Event::instruction_trap());
 
-  t->push_event(Event::instruction_trap());
+  if (t->retry_syscall_patching) {
+    LOG(debug) << "Retrying deferred syscall patching";
+    t->retry_syscall_patching = false;
+    if (t->vm()->monkeypatcher().try_patch_trapping_instruction(t, len, false)) {
+      // Instruction was patched. Emit event.
+      auto ev = Event::patch_syscall();
+      ev.PatchSyscall().patch_after_syscall = true;
+      t->record_event(ev);
+    }
+  }
+
+  t->push_event(Event::noop());
   return true;
 }
 
@@ -438,7 +449,7 @@ static void handle_desched_event(RecordTask* t) {
    *  o parent is notified and sees counter value i+1
    *  o parent stops delivery of first signal and disarms
    *    counter
-   *  o second signal dequeued and delivered, notififying parent
+   *  o second signal dequeued and delivered, notifying parent
    *    (counter is disarmed now, so no pseudo-desched possible
    *    here)
    *  o parent notifiedand sees counter value i+1 again
@@ -468,7 +479,7 @@ static void handle_desched_event(RecordTask* t) {
       // Untraced syscall, we may not have set original_syscallno for this on aarch64.
       syscallno = t->regs().syscallno();
     } else {
-      // On x86, syscall no is overwriten by return value.
+      // On x86, syscall no is overwritten by return value.
       ASSERT(t, is_x86ish(t->arch()));
       syscallno = t->regs().original_syscallno();
     }
@@ -507,7 +518,11 @@ static void handle_desched_event(RecordTask* t) {
     // syscall may have re-armed the event.
     disarm_desched_event(t);
 
-    t->resume_execution(RESUME_SYSCALL, RESUME_WAIT, RESUME_UNLIMITED_TICKS);
+    if (!t->resume_execution(RESUME_SYSCALL, RESUME_WAIT_NO_EXIT, RESUME_UNLIMITED_TICKS)) {
+      LOG(debug) << "  (got exit, bailing out)";
+      t->push_event(Event::noop());
+      return;
+    }
 
     if (t->status().is_syscall()) {
       t->apply_syscall_entry_regs();
@@ -526,12 +541,6 @@ static void handle_desched_event(RecordTask* t) {
       // seccomp filter.
       break;
     }
-    if (t->ptrace_event() == PTRACE_EVENT_EXIT) {
-      LOG(debug)
-          << "  (got exit, bailing out)";
-      t->push_event(Event::noop());
-      return;
-    }
 
     // Completely ignore spurious desched signals and
     // signals that aren't going to be delivered to the
