@@ -218,7 +218,8 @@ template <typename Arch> static void prepare_clone(ReplayTask* t) {
   Registers entry_regs = r;
 
   // Run; we will be interrupted by PTRACE_EVENT_CLONE/FORK/VFORK.
-  t->resume_execution(RESUME_CONT, RESUME_WAIT, RESUME_NO_TICKS);
+  bool ok = t->resume_execution(RESUME_CONT, RESUME_WAIT_NO_EXIT, RESUME_NO_TICKS);
+  ASSERT(t, ok) << "Tracee was killed";
 
   pid_t new_tid;
   while (!t->clone_syscall_is_complete(&new_tid, Arch::arch())) {
@@ -227,7 +228,8 @@ template <typename Arch> static void prepare_clone(ReplayTask* t) {
     // state to try the syscall again.
     ASSERT(t, t->regs().syscall_result_signed() == -EAGAIN);
     t->set_regs(entry_regs);
-    t->resume_execution(RESUME_CONT, RESUME_WAIT, RESUME_NO_TICKS);
+    bool ok = t->resume_execution(RESUME_CONT, RESUME_WAIT_NO_EXIT, RESUME_NO_TICKS);
+    ASSERT(t, ok) << "Tracee was killed";
   }
 
   // Get out of the syscall
@@ -249,7 +251,7 @@ template <typename Arch> static void prepare_clone(ReplayTask* t) {
   t->set_regs(r);
   t->canonicalize_regs(trace_frame.event().Syscall().arch());
 
-  // Dig the recorded tid out out of the trace. The tid value returned in
+  // Dig the recorded tid out of the trace. The tid value returned in
   // the recorded registers could be in a different pid namespace from rr's,
   // so we can't use it directly.
   TraceTaskEvent tte = read_task_trace_event(t, TraceTaskEvent::CLONE);
@@ -311,8 +313,6 @@ template <typename Arch> static void prepare_clone(ReplayTask* t) {
   TraceReader::MappedData data;
   KernelMapping km = t->trace_reader().read_mapped_region(&data);
   init_scratch_memory(new_task, km, data);
-
-  new_task->vm()->after_clone();
 }
 
 static void restore_mapped_region(ReplayTask* t, AutoRemoteSyscalls& remote,
@@ -403,7 +403,7 @@ static void process_execve(ReplayTask* t, const TraceFrame& trace_frame,
       // (up to) two kms that have different fsnames but do each have an
       // executable segment, as well as the km that contains the ip. This is
       // slightly complicated, but should handle the case where either file has
-      // more than one exectuable segment.
+      // more than one executable segment.
       const string& file_name = km.fsname();
       if ((km.prot() & PROT_EXEC) && file_name.size() > 0 &&
           // Make sure to exclude [vdso] (and similar) and executable stacks.
@@ -462,7 +462,7 @@ static void process_execve(ReplayTask* t, const TraceFrame& trace_frame,
 
     // Now fix up the address space. First unmap all the mappings other than
     // our rr page.
-    t->vm()->unmap_all_but_rr_page(remote);
+    t->vm()->unmap_all_but_rr_mappings(remote);
     // We will have unmapped the stack memory that |remote| would have used for
     // memory parameters. Fortunately process_mapped_region below doesn't
     // need any memory parameters for its remote syscalls.
@@ -471,8 +471,6 @@ static void process_execve(ReplayTask* t, const TraceFrame& trace_frame,
     restore_mapped_region(t, remote, kms[0], datas[0]);
   }
 
-  const string& recorded_exe_name = kms[exe_km].fsname();
-
   {
     // Now that [stack] is mapped, reinitialize AutoRemoteSyscalls with
     // memory parameters enabled.
@@ -482,14 +480,6 @@ static void process_execve(ReplayTask* t, const TraceFrame& trace_frame,
     for (ssize_t i = 1; i < ssize_t(kms.size()) - 1; ++i) {
       restore_mapped_region(t, remote, kms[i], datas[i]);
     }
-
-    size_t index = recorded_exe_name.rfind('/');
-    string name =
-        string("rr:") +
-        recorded_exe_name.substr(index == string::npos ? 0 : index + 1);
-    AutoRestoreMem mem(remote, name.c_str());
-    remote.infallible_syscall(syscall_number_for_prctl(t->arch()), PR_SET_NAME,
-                              mem.get());
   }
 
   init_scratch_memory(t, kms.back(), datas.back());
@@ -814,6 +804,7 @@ static void process_mremap(ReplayTask* t, const TraceFrame& trace_frame,
   size_t old_size = ceil_page_size(trace_regs.arg2());
   remote_ptr<void> new_addr = trace_frame.regs().syscall_result();
   size_t new_size = ceil_page_size(trace_regs.arg3());
+  int flags = trace_regs.arg4_signed();
 
   // The recorded mremap call succeeded, so we know the original mapping can be
   // treated as a single mapping.
@@ -845,13 +836,13 @@ static void process_mremap(ReplayTask* t, const TraceFrame& trace_frame,
       // addresses? Hopefully the kernel doesn't do that to us!!!
       remote.infallible_syscall_if_alive(trace_regs.original_syscallno(), old_addr,
                                          old_size, new_size,
-                                         MREMAP_MAYMOVE | MREMAP_FIXED, new_addr);
+                                         flags | MREMAP_MAYMOVE | MREMAP_FIXED, new_addr);
     }
 
     remote.regs().set_syscall_result(new_addr);
   }
 
-  t->vm()->remap(t, old_addr, old_size, new_addr, new_size);
+  t->vm()->remap(t, old_addr, old_size, new_addr, new_size, flags);
 
   AddressSpace::Mapping mapping = t->vm()->mapping_of(new_addr);
   auto f = mapping.emu_file;
@@ -948,18 +939,10 @@ static void process_shmdt(ReplayTask* t, const TraceFrame& trace_frame,
 static void process_init_buffers(ReplayTask* t, ReplayTraceStep* step) {
   step->action = TSTEP_RETIRE;
 
-  /* Proceed to syscall exit so we can run our own syscalls. */
-  remote_ptr<void> rec_child_map_addr =
-      t->current_trace_frame().regs().syscall_result();
-
   /* We don't want the desched event fd during replay, because
    * we already know where they were.  (The perf_event fd is
    * emulated anyway.) */
-  t->init_buffers(rec_child_map_addr);
-
-  ASSERT(t, t->syscallbuf_child.cast<void>() == rec_child_map_addr)
-      << "Should have mapped syscallbuf at " << rec_child_map_addr
-      << ", but it's at " << t->syscallbuf_child;
+  t->init_buffers();
   t->validate_regs();
 }
 
@@ -1199,6 +1182,23 @@ static void rep_process_syscall_arch(ReplayTask* t, ReplayTraceStep* step,
       switch ((int)t->regs().arg3()) {
         case MADV_DONTNEED:
         case MADV_REMOVE:
+        case MADV_DONTNEED_LOCKED:
+          break;
+        /* These are not technically required to be passed through, but the
+           syscallbuf code does, so if we don't here, we risk fracturing
+           otherwise coalescable memory regions. Asan in particular triggers
+           a pathological case here that quickly exhausts the total mapping
+           limit by fracturing its shadow region */
+        case MADV_NORMAL:
+        case MADV_RANDOM:
+        case MADV_SEQUENTIAL:
+        case MADV_WILLNEED:
+        case MADV_MERGEABLE:
+        case MADV_UNMERGEABLE:
+        case MADV_HUGEPAGE:
+        case MADV_NOHUGEPAGE:
+        case MADV_DONTDUMP:
+        case MADV_DODUMP:
           break;
         default:
           return;
