@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: nil; -*- */
 
-#include "rrutil.h"
+#include "util.h"
 
 void catcher(__attribute__((unused)) int signum,
              __attribute__((unused)) siginfo_t* siginfo_ptr,
