@@ -11,9 +11,13 @@ expect_gdb('Breakpoint 1')
 send_gdb('call crash()')
 expect_gdb('SIGSEGV')
 send_gdb('c')
+expect_gdb('Program terminated with signal SIGKILL')
+# gdb forgets that we already stopped at this breakpoint so it'll stop us there again
+send_gdb('next')
 expect_gdb('Breakpoint 1')
 
 send_gdb('next')
+expect_rr('EXIT-SUCCESS')
 send_gdb('call crash()')
 expect_gdb('SIGSEGV')
 send_gdb('c')
@@ -27,6 +31,7 @@ send_gdb('call crash()')
 expect_gdb('SIGSEGV')
 
 send_gdb('next')
+expect_rr('EXIT-SUCCESS')
 send_gdb('call crash()')
 expect_gdb('SIGSEGV')
 send_gdb('c')
