@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: nil; -*- */
 
-#include "rrutil.h"
+#include "util.h"
 
 static int caught_signal;
 static void handle_signal(__attribute__((unused)) int sig) { ++caught_signal; }
