@@ -6,6 +6,7 @@
 
 int main(void) {
   int fd = open(DEV_RANDOM, O_RDONLY);
+  int entropy_count;
   char buf[128];
   ssize_t nread;
 
@@ -17,6 +18,9 @@ int main(void) {
 
   check_data(buf, sizeof(buf));
 
+  test_assert(ioctl(fd, RNDGETENTCNT, &entropy_count) == 0);
+  atomic_printf("Entropy Count: %d\n", entropy_count);
+
   atomic_puts("EXIT-SUCCESS");
   return 0;
 }
