@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: nil; -*- */
 
-#include "rrutil.h"
+#include "util.h"
 
 static void queue_siginfo(int sig, int val) {
   siginfo_t si;
@@ -35,7 +35,7 @@ static void handle_signal(int sig, siginfo_t* si,
   } else if (SIGUSR2 == sig) {
     usr2_val = val;
   } else {
-    assert("Unexpected signal" && 0);
+    test_assert("Unexpected signal" && 0);
   }
 }
 
