@@ -6,7 +6,7 @@
 
 extern int capset(cap_user_header_t header, const cap_user_data_t data);
 
-static char tmp_name[] = "temp";
+static char tmp_name[] = "tempXXXXXX";
 static uid_t uid;
 static uid_t gid;
 
@@ -29,51 +29,51 @@ static void test_mount(void) {
 }
 
 static void test_uids(void) {
-  test_assert(0 == syscall(SYS_setreuid, 7, 7));
+  test_assert(0 == syscall(SYS_setreuid, 7, 7) || errno == ENOSYS);
 #ifdef SYS_setreuid32
   test_assert(0 == syscall(SYS_setreuid32, 7, 7));
 #endif
-  test_assert(0 == syscall(SYS_setregid, 8, 8));
+  test_assert(0 == syscall(SYS_setregid, 8, 8) || errno == ENOSYS);
 #ifdef SYS_setreuid32
   test_assert(0 == syscall(SYS_setregid32, 8, 8));
 #endif
-  test_assert(0 == syscall(SYS_setresuid, 7, 7, 7));
+  test_assert(0 == syscall(SYS_setresuid, 7, 7, 7) || errno == ENOSYS);
 #ifdef SYS_setresuid32
   test_assert(0 == syscall(SYS_setresuid32, 7, 7, 7));
 #endif
-  test_assert(0 == syscall(SYS_setresgid, 8, 8, 8));
+  test_assert(0 == syscall(SYS_setresgid, 8, 8, 8) || errno == ENOSYS);
 #ifdef SYS_setresuid32
   test_assert(0 == syscall(SYS_setresgid32, 8, 8, 8));
 #endif
-  test_assert(0 == syscall(SYS_setuid, 7));
+  test_assert(0 == syscall(SYS_setuid, 7) || errno == ENOSYS);
 #ifdef SYS_setuid32
   test_assert(0 == syscall(SYS_setuid32, 7));
 #endif
-  test_assert(0 == syscall(SYS_setgid, 8));
+  test_assert(0 == syscall(SYS_setgid, 8) || errno == ENOSYS);
 #ifdef SYS_setgid32
   test_assert(0 == syscall(SYS_setgid32, 8));
 #endif
-  test_assert(7 == syscall(SYS_setfsuid, 7));
+  test_assert(7 == syscall(SYS_setfsuid, 7) || errno == ENOSYS);
 #ifdef SYS_setfsuid32
   test_assert(7 == syscall(SYS_setfsuid32, 7));
 #endif
-  test_assert(8 == syscall(SYS_setfsgid, 8));
+  test_assert(8 == syscall(SYS_setfsgid, 8) || errno == ENOSYS);
 #ifdef SYS_setfsgid32
   test_assert(8 == syscall(SYS_setfsgid32, 8));
 #endif
-  test_assert(7 == syscall(SYS_getuid));
+  test_assert(7 == syscall(SYS_getuid) || errno == ENOSYS);
 #ifdef SYS_getuid32
   test_assert(7 == syscall(SYS_getuid32));
 #endif
-  test_assert(8 == syscall(SYS_getgid));
+  test_assert(8 == syscall(SYS_getgid) || errno == ENOSYS);
 #ifdef SYS_getgid32
   test_assert(8 == syscall(SYS_getgid32));
 #endif
-  test_assert(7 == syscall(SYS_geteuid));
+  test_assert(7 == syscall(SYS_geteuid) || errno == ENOSYS);
 #ifdef SYS_geteuid32
   test_assert(7 == syscall(SYS_geteuid32));
 #endif
-  test_assert(8 == syscall(SYS_getegid));
+  test_assert(8 == syscall(SYS_getegid) || errno == ENOSYS);
 #ifdef SYS_getegid32
   test_assert(8 == syscall(SYS_getegid32));
 #endif
@@ -127,7 +127,7 @@ static void write_user_namespace_mappings(void) {
   char buf[100];
   int fd;
 
-  fd = open("/proc/self/uid_map", O_WRONLY | O_CREAT);
+  fd = open("/proc/self/uid_map", O_WRONLY | O_CREAT, 0600);
   test_assert(fd >= 0);
   sprintf(buf, "7 %d 1\n", uid);
   test_assert((ssize_t)strlen(buf) == write(fd, buf, strlen(buf)));
@@ -136,7 +136,7 @@ static void write_user_namespace_mappings(void) {
   /* Per user_namespaces(7), we need to write 'deny' to /proc/self/setgroups
    * before we set up gid_map. This means we can't test setgroups here. Oh well.
    */
-  fd = open("/proc/self/setgroups", O_WRONLY | O_CREAT);
+  fd = open("/proc/self/setgroups", O_WRONLY | O_CREAT, 0600);
   /* This file does not exist in some old kernel */
   if (fd >= 0) {
     sprintf(buf, "deny");
@@ -144,7 +144,7 @@ static void write_user_namespace_mappings(void) {
     test_assert(0 == close(fd));
   }
 
-  fd = open("/proc/self/gid_map", O_WRONLY | O_CREAT);
+  fd = open("/proc/self/gid_map", O_WRONLY | O_CREAT, 0600);
   test_assert(fd >= 0);
   sprintf(buf, "8 %d 1\n", gid);
   test_assert((ssize_t)strlen(buf) == write(fd, buf, strlen(buf)));
@@ -245,7 +245,7 @@ int main(void) {
   uid = getuid();
   gid = getgid();
 
-  test_assert(0 == mkdir(tmp_name, 0700));
+  mkdtemp(tmp_name);
 
   child = fork();
   if (!child) {
