@@ -4,7 +4,7 @@ import re
 send_gdb('handle SIGKILL stop')
 
 send_gdb('c')
-expect_gdb('SIGKILL')
+expect_gdb('Program received signal SIGKILL')
 
 send_gdb('rc')
 expect_gdb('Program stopped.')
