@@ -5,7 +5,7 @@
 #include <arpa/inet.h>
 #include <dirent.h>
 #include <elf.h>
-#ifdef EXECINFO_H
+#ifdef EXECINFO_BACKTRACE
 #include <execinfo.h>
 #endif
 #include <fcntl.h>
@@ -43,6 +43,7 @@
 #include "RecordTask.h"
 #include "ReplayTask.h"
 #include "TraceStream.h"
+#include "WaitManager.h"
 #include "core.h"
 #include "kernel_abi.h"
 #include "kernel_metadata.h"
@@ -446,11 +447,10 @@ static void iterate_checksums(Task* t, ChecksumMode mode,
     FATAL() << "Failed to open checksum file " << filename;
   }
 
+  ReplaySession *replay = t->session().as_replay();
   remote_ptr<unsigned char> in_replay_flag;
-  unsigned char in_replay = 0;
-  if (t->preload_globals) {
-    in_replay_flag = REMOTE_PTR_FIELD(t->preload_globals, in_replay);
-    in_replay = t->read_mem(in_replay_flag);
+  if (replay && replay->has_trace_quirk(TraceReader::UsesGlobalsInReplay) && t->preload_globals) {
+    in_replay_flag = REMOTE_PTR_FIELD(t->preload_globals, reserved_legacy_in_replay);
     t->write_mem(in_replay_flag, (unsigned char)0);
   }
 
@@ -552,10 +552,18 @@ static void iterate_checksums(Task* t, ChecksumMode mode,
       * the "extra data" for the possibly one pending
       * record.
       *
+      * The deterministic region excludes the notify_on_syscall_hook_exit
+      * flag. This flag is written by is_safe_to_deliver_signal and
+      * that write can occur at a different event to where ReplaySession
+      * eventually sets it.
+      *
       * So here, we set things up so that we only checksum
       * the deterministic region. */
-      auto child_hdr = m.map.start().cast<struct syscallbuf_hdr>();
-      auto hdr = t->read_mem(child_hdr);
+      struct syscallbuf_hdr hdr;
+      ASSERT(t, mem.size() >= sizeof(hdr));
+      memcpy(&hdr, mem.data(), sizeof(hdr));
+      hdr.notify_on_syscall_hook_exit = 0;
+      memcpy(mem.data(), &hdr, sizeof(hdr));
       mem.resize(sizeof(hdr) + hdr.num_rec_bytes +
                  sizeof(struct syscallbuf_record));
     }
@@ -577,7 +585,7 @@ static void iterate_checksums(Task* t, ChecksumMode mode,
   }
 
   if (in_replay_flag) {
-    t->write_mem(in_replay_flag, in_replay);
+    t->write_mem(in_replay_flag, (unsigned char)1);
   }
 
   fclose(c.checksums_file);
@@ -799,7 +807,7 @@ bool should_copy_mmap_region(const KernelMapping& mapping,
      * frequent than even system updates.
      *
      * XXX what about the fontconfig cache files? */
-    LOG(debug) << "  (no copy for root-owned " << file_name << ")";
+    LOG(debug) << "  (no copy for root-owned " << mapping.fsname() << ")";
     return false;
   }
   if (private_mapping) {
@@ -1739,7 +1747,7 @@ void notifying_abort() {
 void dump_rr_stack() {
   static const char msg[] = "=== Start rr backtrace:\n";
   write_all(STDERR_FILENO, msg, sizeof(msg) - 1);
-#if EXECINFO_H
+#ifdef EXECINFO_BACKTRACE
   void* buffer[1024];
   int count = backtrace(buffer, 1024);
   backtrace_symbols_fd(buffer, count, STDERR_FILENO);
@@ -1766,11 +1774,12 @@ void check_for_leaks() {
 }
 
 void ensure_dir(const string& dir, const char* dir_type, mode_t mode) {
-  string d = dir;
-  while (!d.empty() && d[d.length() - 1] == '/') {
-    d = d.substr(0, d.length() - 1);
+  if (dir.empty()) {
+    FATAL() << "Empty directory name!";
   }
-
+  size_t last_dir_component = dir.find_last_not_of('/');
+  const string d = last_dir_component == string::npos
+    ? "/" : dir.substr(0, last_dir_component + 1);
   struct stat st;
   if (0 > stat(d.c_str(), &st)) {
     if (errno != ENOENT) {
@@ -1778,10 +1787,9 @@ void ensure_dir(const string& dir, const char* dir_type, mode_t mode) {
     }
 
     size_t last_slash = d.find_last_of('/');
-    if (last_slash == string::npos || last_slash == 0) {
-      FATAL() << "Can't find directory `" << dir << "'";
+    if (last_slash != string::npos) {
+      ensure_dir(d.substr(0, max(last_slash, (size_t)1)), dir_type, mode);
     }
-    ensure_dir(d.substr(0, last_slash), dir_type, mode);
 
     // Allow for a race condition where someone else creates the directory
     if (0 > mkdir(d.c_str(), mode) && errno != EEXIST) {
@@ -1790,6 +1798,7 @@ void ensure_dir(const string& dir, const char* dir_type, mode_t mode) {
     if (0 > stat(d.c_str(), &st)) {
       FATAL() << "Can't stat " << dir_type << " `" << dir << "'";
     }
+    return;
   }
 
   if (!(S_IFDIR & st.st_mode)) {
@@ -1986,7 +1995,7 @@ bool is_advanced_pc_and_signaled_instruction(Task* t, remote_code_ptr ip) {
   return false;
 }
 
-static string get_cpu_lock_file() {
+string get_cpu_lock_file() {
   const char* lock_file = getenv("_RR_CPU_LOCK_FILE");
   return lock_file ? lock_file : trace_save_dir() + "/cpu_lock";
 }
@@ -2191,6 +2200,73 @@ bool is_directory(const char* path) {
   return (buf.st_mode & S_IFDIR) != 0;
 }
 
+const char* filename(const char* path) {
+  const char* dir = strrchr(path, '/');
+  return dir ? ++dir : path;
+}
+
+bool is_trace(const string& trace) {
+  string ver = trace + "/version";
+  string inc = trace + "/incomplete";
+  return access(ver.c_str(), F_OK) == 0 || access(inc.c_str(), F_OK) == 0;
+}
+
+bool is_latest_trace(const string& trace) {
+  string latest = latest_trace_symlink();
+  if (access(latest.c_str(), F_OK) != 0) {
+    return false;
+  }
+  latest = real_path(latest);
+  return latest == trace; 
+}
+
+bool remove_latest_trace_symlink() {
+  const string latest = latest_trace_symlink();
+  int ret = remove(latest.c_str());
+  if (ret) {
+    perror(latest.c_str());
+    fprintf(stderr,
+            "\n"
+            "rr: Failed to remove latest_trace symlink: error code %d\n"
+            "\n",
+            ret);
+    return false;
+  }
+  return true;
+}
+
+bool is_valid_trace_name(const string& entry, std::string* reason) {
+  // filename corresponds to dirname
+  const string name = filename(entry.c_str());
+
+  if (name.empty()) {
+    if (reason) {
+      *reason = "Empty";
+    }
+    return false;
+  }
+  if (name[0] == '.' || name[0] == '#') {
+    if (reason) {
+      *reason = "Cannot start with . or #";
+    }
+    return false;
+  }
+  if (name[name.length() - 1] == '~') {
+    if (reason) {
+      *reason = "Cannot end with ~";
+    }
+    return false;
+  }
+  if (name == "cpu_lock") {
+    if (reason) {
+      *reason = "Name cpu_lock is reserved";
+    }
+    return false;
+  }
+
+  return true;
+}
+
 ssize_t read_to_end(const ScopedFd& fd, size_t offset, void* buf, size_t size) {
   ssize_t ret = 0;
   while (size) {
@@ -2378,6 +2454,15 @@ static int child_SIGSEGV(__attribute__((unused)) void* arg) {
   return 0;
 }
 
+#ifdef __SANITIZE_ADDRESS__
+#include <sanitizer/lsan_interface.h>
+static int rr_lsan_is_turned_off;
+int __lsan_is_turned_off(void)
+{
+  return rr_lsan_is_turned_off;
+}
+#endif
+
 bool coredumping_signal_takes_down_entire_vm() {
   // The kernel behavior here changed in 5.16. Prior to that,
   // a coredumping signal would bring down the entire vm.
@@ -2386,14 +2471,16 @@ bool coredumping_signal_takes_down_entire_vm() {
   // 1 the old behavior.
   static int coredumping_signal_vm_behavior = -1;
   if (coredumping_signal_vm_behavior < 0) {
-    pid_t child = -1;
-    int status = -1;
+    pid_t child;
     LOG(debug) << "Testing coredumping behavior in the presence of CLONE_VM";
     if ((child = fork()) == 0) {
       // Remove any handlers and make sure we get the default signal behavior.
       signal(SIGSEGV, SIG_DFL);
       // Don't litter the system with core dumps.
       prctl(PR_SET_DUMPABLE, 0);
+#ifdef __SANITIZE_ADDRESS__
+      rr_lsan_is_turned_off = 1;
+#endif
       // Allocate a stack for the child.
       const size_t stack_size = 1 << 20;
       void* stack = mmap(NULL, stack_size, PROT_READ | PROT_WRITE,
@@ -2403,16 +2490,16 @@ bool coredumping_signal_takes_down_entire_vm() {
                       CLONE_VM, NULL, NULL, NULL, NULL);
       DEBUG_ASSERT(tid > 0);
 
-      pid_t ret = waitpid(tid, &status, __WALL);
-      DEBUG_ASSERT(ret == tid);
-      DEBUG_ASSERT(WIFSIGNALED(status));
+      WaitResult result = WaitManager::wait_exit(WaitOptions(tid));
+      DEBUG_ASSERT(result.code == WAIT_OK);
+      DEBUG_ASSERT(result.status.type() == WaitStatus::FATAL_SIGNAL);
       exit(0);
     }
 
     DEBUG_ASSERT(child > 0);
-    pid_t ret = waitpid(child, &status, __WALL);
-    DEBUG_ASSERT(ret == child);
-    coredumping_signal_vm_behavior = WIFSIGNALED(status);
+    WaitResult result = WaitManager::wait_exit(WaitOptions(child));
+    DEBUG_ASSERT(result.code == WAIT_OK);
+    coredumping_signal_vm_behavior = result.status.type() == WaitStatus::FATAL_SIGNAL;
   }
 
   return coredumping_signal_vm_behavior > 0;
