@@ -490,6 +490,9 @@ bool is_advanced_pc_and_signaled_instruction(Task* t, remote_code_ptr ip);
  */
 enum BindCPU { BIND_CPU = -2, UNBOUND_CPU = -1 };
 
+/* Get the path of the cpu lock file */
+std::string get_cpu_lock_file();
+
 /* Convert a BindCPU to a specific CPU number. If possible, the cpu_lock_fd_out
    will be set to an fd that holds an advisory fcntl lock for the chosen CPU
    for coordination with other rr processes */
@@ -512,6 +515,38 @@ ssize_t pwrite_all_fallible(int fd, const void* buf, size_t size, off64_t offset
  */
 bool is_directory(const char* path);
 
+/*
+ * Returns a pointer to the filename portion of the path.
+ * That is the position after the last '/'
+ */
+const char* filename(const char* path);
+
+/*
+ * Returns whether a trace is at the path by checking for a version or
+ * incomplete file.
+ * Will set errno, if false.
+ */
+bool is_trace(const std::string& path);
+
+/*
+ * Returns whether the latest_trace symlink (if any) points to |trace|.
+ */
+bool is_latest_trace(const std::string& trace);
+
+/*
+ * Deletes the latest_trace symlink, logs an error and returns false on failure.
+ */
+bool remove_latest_trace_symlink();
+
+/*
+ * Returns whether |entry| is a valid trace name.
+ * If invalid, optional out-param |reason| will be set to the reason.
+ * I.e. does not start with . or #, does not end with ~, is neither cpu_lock
+ * nor latest_trace.
+ */
+bool is_valid_trace_name(const std::string& entry,
+                         std::string* reason = nullptr);
+
 /**
  * Read bytes from `fd` into `buf` from `offset` until the read returns an
  * error or 0 or the buffer is full. Returns total bytes read or -1 for error.
@@ -584,6 +619,17 @@ inline unsigned long long rdtsc(void) {
 #endif
 }
 
+inline unsigned long long dczid_el0_block_size(void) {
+#if defined(__aarch64__)
+  unsigned long long val;
+  asm volatile("mrs %0, DCZID_EL0" : "=r" (val));
+  return 1ULL << (val & 0xF);
+#else
+  FATAL() << "Reached AArch64-only code path on non-AArch64 architecture";
+  return 0;
+#endif
+}
+
 } // namespace rr
 
 #endif /* RR_UTIL_H_ */
